import json

from qds_sdk.qubole import Qubole
from qds_sdk.resource import Resource
from argparse import ArgumentParser

class GroupCmdLine:
    """
    qds_sdk.GroupCmdLine is the interface used by qds.py.
    """

    @staticmethod
    def parsers():
        argparser = ArgumentParser(prog="qds.py group",
                                        description="Client to manage Groups in Qubole Data Service.")
        subparsers = argparser.add_subparsers()

        #Create
        create = subparsers.add_parser("create", help="Creates a new Group")
        create.add_argument("--name", dest="name", required=True, help="Name of the new Group.")
        create.add_argument("--members", dest="members", help="List of User ids to be added in this new Group.")
        create.add_argument("--roles", dest="roles", help="List of Role ids to be attached with this new Group.")
        create.set_defaults(func=GroupCmdLine.create)

        #List
        list = subparsers.add_parser("list", help="List all Groups")
        list.add_argument("--per-page", dest="per_page", help="Number of items per page")
        list.add_argument("--page", dest="page", help="Page Number")
        list.set_defaults(func=GroupCmdLine.list)

        #View
        view = subparsers.add_parser("view", help="View a specific Group")
        view.add_argument("id", help="Numeric id of the Group")
        view.set_defaults(func=GroupCmdLine.view)

        #Update
        update = subparsers.add_parser("update", help="Update a specific Group's name")
        update.add_argument("--name", dest="name", help="New name of the Group")
        update.add_argument("--members", dest="members", help="List of User ids to be added in this Group")
        update.add_argument("--roles", dest="roles", help="List of Role ids to be attached with this Group.")
        update.add_argument("--remove-members", dest="remove_members",
                            help="List of User ids to be removed from this Group")
        update.add_argument("--remove-roles", dest="remove_roles",
                            help="List of Role ids to be detached from this Group")
        update.add_argument("id", help="Numeric id of the Group")
        update.set_defaults(func=GroupCmdLine.update)

        #Delete
        delete = subparsers.add_parser("delete", help="Delete a Group")
        delete.add_argument("id", help="Numeric id of the Group to be deleted")
        delete.set_defaults(func=GroupCmdLine.delete)

        #duplicate
        duplicate = subparsers.add_parser("duplicate", help="Duplicates/Clones a group")
        duplicate.add_argument("id", help="Numeric id of the Group to be Cloned")
        duplicate.add_argument("--name", dest="name", required=False, help="Name of the new group")
        duplicate.set_defaults(func=GroupCmdLine.duplicate)

        #Add user
        add_user = subparsers.add_parser("add-users", help="Add Users to the Group")
        add_user.add_argument("id", help="Numeric id of the group")
        add_user.add_argument("user_ids", help="User IDs of the Users to be added in this Group")
        add_user.set_defaults(func=GroupCmdLine.add_users)

        #Remove user
        remove_user = subparsers.add_parser("remove-users", help="Remove Users from the Group")
        remove_user.add_argument("id", help="Numeric id of the group")
        remove_user.add_argument("user_ids", help="User IDs of the Users to be removed from this Group")
        remove_user.set_defaults(func=GroupCmdLine.remove_users)

        #List roles for a group
        list_roles = subparsers.add_parser("list-roles", help="List all Roles of a Group ")
        list_roles.add_argument("id", help="Numeric id of the group")
        list_roles.set_defaults(func=GroupCmdLine.list_roles)

        #Add role
        add_role = subparsers.add_parser("add-roles", help="Attach Roles to the Group")
        add_role.add_argument("id", help="Numeric id of the group")
        add_role.add_argument("role_id", help="Please provide the Role IDs, which you likes to attach with this Group.")
        add_role.set_defaults(func=GroupCmdLine.add_roles)

        #Remove role
        remove_role = subparsers.add_parser("remove-roles", help="Detach Roles from the Group")
        remove_role.add_argument("id", help="Numeric id of the group")
        remove_role.add_argument("role_id", help="Please provide the Role IDs, which you likes to detach from this Group.")
        remove_role.set_defaults(func=GroupCmdLine.remove_roles)

        #List users for a group
        list_users = subparsers.add_parser("list-users", help="List all Users of a Group ")
        list_users.add_argument("id", help="Numeric id of the group")
        list_users.set_defaults(func=GroupCmdLine.list_users)

        return argparser

    @staticmethod
    def run(args):
        parser = GroupCmdLine.parsers()
        parsed = parser.parse_args(args)
        return parsed.func(parsed)

    @staticmethod
    def create(args):
        group = Group.create(name=args.name, members=args.members, roles=args.roles)
        return json.dumps(group.attributes, sort_keys=True, indent=4)

    @staticmethod
    def list(args):
        grouplist = Group.list(args.page, args.per_page)
        return json.dumps(grouplist, default=lambda o: o.attributes, sort_keys=True, indent=4)

    @staticmethod
    def view(args):
        group = Group.find(args.id)
        return json.dumps(group.attributes, sort_keys=True, indent=4)

    @staticmethod
    def update(args):
        options = {}
        if args.name is not None:
          options["name"] = args.name
        if args.members is not None:
          options["members"] = args.members
        if args.roles is not None:
          options["roles"] = args.roles
        if args.remove_members is not None:
          options["removed_members"] = args.remove_members
        if args.remove_roles is not None:
          options["removed_roles"] = args.remove_roles
        return json.dumps(Group.update(args.id, **options), sort_keys=True, indent=4)

    @staticmethod
    def delete(args):
        return json.dumps(Group.delete(args.id), sort_keys=True, indent=4)

    @staticmethod
    def add_users(args):
        options = {}
        if args.user_ids is not None:
            options["members"] = args.user_ids
        return json.dumps(Group.update(args.id, **options), sort_keys=True, indent=4)

    @staticmethod
    def remove_users(args):
        options = {}
        if args.user_ids is not None:
            options["removed_members"] = args.user_ids
        return json.dumps(Group.update(args.id, **options), sort_keys=True, indent=4)

    @staticmethod
    def duplicate(args):
        options = {}
        if args.name is not None:
          options["name"] = args.name
        return json.dumps(Group.duplicate(args.id, **options), sort_keys=True, indent=4)

    @staticmethod
    def list_roles(args):
        return json.dumps(Group.list_roles(args.id), sort_keys=True, indent=4)

    @staticmethod
    def list_users(args):
        return json.dumps(Group.list_users(args.id), sort_keys=True, indent=4)

    @staticmethod
    def add_roles(args):
        options = {}
        if args.role_id is not None:
            options["roles"] = args.role_id
        return json.dumps(Group.update(args.id, **options), sort_keys=True, indent=4)

    @staticmethod
    def remove_roles(args):
        options = {}
        if args.role_id is not None:
            options["removed_roles"] = args.role_id
        return json.dumps(Group.update(args.id, **options), sort_keys=True, indent=4)



class Group(Resource):
    """
    qds_sdk.Group is the base Qubole Group class.
    """

    """ all commands use the /group endpoint"""

    rest_entity_path = "groups"

    @staticmethod
    def list(page = None, per_page = None):
        conn = Qubole.agent()
        url_path = Group.rest_entity_path
        page_attr = []
        if page is not None:
            page_attr.append("page=%s" % page)
        if per_page is not None:
            page_attr.append("per_page=%s" % per_page)
        if page_attr:
            url_path = "%s?%s" % (Group.rest_entity_path, "&".join(page_attr))

        groupjson = conn.get(url_path)
        grouplist = []
        for s in groupjson["groups"]:
            grouplist.append(Group(s))
        return grouplist

    @staticmethod
    def update(group_id, **kwargs):
        conn = Qubole.agent()
        url_path = "groups/%s" % group_id
        return conn.put(url_path, data=kwargs)

    @staticmethod
    def delete(group_id):
        conn = Qubole.agent()
        url_path = "groups/%s" % group_id
        return conn.delete(url_path)

    @staticmethod
    def add_user(group_id, user_id):
        conn = Qubole.agent()
        url_path = "groups/%s/qbol_users/%s/add" % (group_id, user_id)
        return conn.put(url_path)

    @staticmethod
    def remove_user(group_id, user_id):
        conn = Qubole.agent()
        url_path = "groups/%s/qbol_users/%s/remove" % (group_id, user_id)
        return conn.put(url_path)

    @staticmethod
    def duplicate(group_id, **kwargs):
        conn = Qubole.agent()
        url_path = "groups/%s/duplicate" % group_id
        return conn.post(url_path, data=kwargs)

    @staticmethod
    def list_roles(group_id):
        conn = Qubole.agent()
        url_path = "groups/%s/roles" % group_id
        return conn.get(url_path)

    @staticmethod
    def list_users(group_id):
        conn = Qubole.agent()
        url_path = "groups/%s/qbol_users" % group_id
        return conn.get(url_path)





import json

from qds_sdk.qubole import Qubole
from qds_sdk.resource import Resource
from argparse import ArgumentParser
from qds_sdk.commands import *

class DbTapCmdLine:
    """
    qds_sdk.ScheduleCmdLine is the interface used by qds.py.
    """

    @staticmethod
    def parsers():
        argparser = ArgumentParser(prog="qds.py dbtaps",
                                        description="DbTaps client for Qubole Data Service.")
        subparsers = argparser.add_subparsers()

        #Create
        create = subparsers.add_parser("create",
                                       help="Create a new schedule")
        create.add_argument("--name", dest="name", required=True,
                            help="Database Name")
        create.add_argument("--host", dest="host", required=True,
                            help="Host name or IP address of the database")
        create.add_argument("--user", dest="user", required=True,
                            help="Username")
        create.add_argument("--password", dest="password", required=True,
                            help="Password")
        create.add_argument("--port", dest="port",
                            help="Database Port")
        create.add_argument("--type", dest="type", choices=["mysql","vertica","mongo","postgresql","redshift","sqlserver"],
                            help="Type of database")
        create.add_argument("--location", dest="location", choices=["us-east-1", "us-west-2", "ap-southeast-1", "eu-west-1", "on-premise"],
                            help="Type of database", required=True)
        create.set_defaults(func=DbTapCmdLine.create)

        #List
        list = subparsers.add_parser("list",
                                     help="List all DbTaps")
        list.add_argument("--fields", nargs="*", dest="fields",
                          help="List of fields to show")
        list.add_argument("--per-page", dest="per_page",
                          help="Number of items per page")
        list.add_argument("--page", dest="page",
                          help="Page Number")
        list.set_defaults(func=DbTapCmdLine.list)

        #View
        view = subparsers.add_parser("view",
                                     help="View a specific DbTap")
        view.add_argument("id", help="Numeric id or name of the DbTap")
        view.add_argument("--fields", nargs="*", dest="fields",
                          help="List of fields to show")
        view.set_defaults(func=DbTapCmdLine.view)

        #Tables
        tables = subparsers.add_parser("tables",
                                        help="List tables in a DbTap")
        tables.add_argument("id", help="Numeric id of the DbTap")
        tables.set_defaults(func=DbTapCmdLine.tables)

        #Edit
        edit = subparsers.add_parser("edit",
                                       help="Edit a DbTap")
        edit.add_argument("id", help="Numeric id of the schedule")
        edit.add_argument("--name", dest="name",
                            help="Database Name")
        edit.add_argument("--host", dest="host",
                            help="Host name or IP address of the database")
        edit.add_argument("--user", dest="user",
                            help="Username")
        edit.add_argument("--password", dest="password",
                            help="Password")
        edit.add_argument("--port", dest="port",
                            help="Database Port")
        edit.add_argument("--type", dest="type", choices=["mysql","vertica","mongo","postgresql","redshift","sqlserver"],
                            help="Type of database")
        edit.add_argument("--location", dest="location", choices=["us-east-1", "us-west-2", "ap-southeast-1", "eu-west-1", "on-premise"],
                            help="Type of database")
        edit.set_defaults(func=DbTapCmdLine.edit)

        #Kill
        kill = subparsers.add_parser("delete",
                                     help="Delete a DbTap")
        kill.add_argument("id", help="Numeric id of the schedule")
        kill.set_defaults(func=DbTapCmdLine.kill)
        return argparser

    @staticmethod
    def run(args):
        parser = DbTapCmdLine.parsers()
        parsed = parser.parse_args(args)
        return parsed.func(parsed)

    @staticmethod
    def filter_fields(schedule, fields):
        filtered = {}
        for field in fields:
            filtered[field] = schedule[field]
        return filtered

    @staticmethod
    def create(args):
        dbtap = DbTap.create(db_name=args.name,
                             db_host=args.host,
                             db_user=args.user,
                             db_passwd=args.password,
                             db_type=args.type,
                             db_location=args.location)

        return json.dumps(dbtap.attributes, sort_keys=True, indent=4)

    @staticmethod
    def list(args):
        taplist = DbTap.list(args.page, args.per_page)
        if args.fields:
            for s in taplist:
                s.attributes = DbTapCmdLine.filter_fields(s.attributes, args.fields)
        return json.dumps(taplist, default=lambda o: o.attributes, sort_keys=True, indent=4)

    @staticmethod
    def view(args):
        tap = DbTap.find(args.id)
        if args.fields:
            tap.attributes = DbTapCmdLine.filter_fields(tap.attributes, args.fields)
        return json.dumps(tap.attributes, sort_keys=True, indent=4)

    @staticmethod
    def tables(args):
        tap = DbTap.find(args.id)
        return json.dumps(tap.tables(), sort_keys=True, indent=4)

    @staticmethod
    def edit(args):
        tap = DbTap.find(args.id)
        """ Carefully setup a dict """
        options = {}
        if not args.name is None:
            options["db_name"]=args.name
        if args.host is not None:
            options["db_host"]=args.host
        if args.user is not None:
            options["db_user"]=args.user
        if args.password is not None:
            options["db_passwd"] = args.password
        if args.type is not None:
            options["db_type"] = args.type
        if args.location is not None:
            options["db_location"] = args.location
        tap = tap.edit(**options)
        return json.dumps(tap.attributes, sort_keys=True, indent=4)

    @staticmethod
    def kill(args):
        tap = DbTap.find(args.id)
        return json.dumps(tap.delete(), sort_keys=True, indent=4)

class DbTap(Resource):
    """
    qds_sdk.Dbtap is the base Qubole DbTap class.
    """

    """ all commands use the /db_taps endpoint"""

    rest_entity_path = "db_taps"

    @staticmethod
    def list(page = None, per_page = None):
        conn = Qubole.agent()
        url_path = DbTap.rest_entity_path
        page_attr = []
        if page is not None:
            page_attr.append("page=%s" % page)
        if per_page is not None:
            page_attr.append("per_page=%s" % per_page)
        if page_attr:
            url_path = "%s?%s" % (DbTap.rest_entity_path, "&".join(page_attr))

        #Todo Page numbers are thrown away right now
        tapjson = conn.get(url_path)
        taplist = []
        for s in tapjson["db_taps"]:
            taplist.append(DbTap(s))
        return taplist

    def tables(self):
        conn = Qubole.agent()
        return conn.get("%s/tables" % self.element_path(self.id))

    def edit(self, **kwargs):
        conn = Qubole.agent()
        return DbTap(conn.put(self.element_path(self.id), data=kwargs))

    def delete(self):
        conn = Qubole.agent()
        return conn.delete(self.element_path(self.id))



import time
import logging
from functools import wraps

log = logging.getLogger("retry")


def retry(ExceptionToCheck, tries=4, delay=3, backoff=2):
    def deco_retry(f):
        @wraps(f)
        def f_retry(*args, **kwargs):
            mtries, mdelay = tries, delay
            while mtries > 1:
                try:
                    return f(*args, **kwargs)
                except ExceptionToCheck as e:
                    msg = "%s, Retrying in %d seconds..." % (e.__class__.__name__, mdelay)
                    log.info(msg)
                    time.sleep(mdelay)
                    mtries -= 1
                    mdelay *= backoff
            return f(*args, **kwargs)
        return f_retry  # true decorator
    return deco_retry



import sys
import requests
import logging
import ssl
import json
import pkg_resources
from requests.adapters import HTTPAdapter
try:
    from requests.packages.urllib3.poolmanager import PoolManager
except ImportError:
    from urllib3.poolmanager import PoolManager
from qds_sdk.retry import retry
from qds_sdk.exception import *


log = logging.getLogger("qds_connection")

"""
see http://stackoverflow.com/questions/14102416/python-requests-requests-exceptions-sslerror-errno-8-ssl-c504-eof-occurred
"""


class MyAdapter(HTTPAdapter):
    def init_poolmanager(self, connections, maxsize,
                         block=False):
        self.poolmanager = PoolManager(num_pools=connections,
                                       maxsize=maxsize,
                                       block=block,
                                       ssl_version=ssl.PROTOCOL_TLSv1)


class Connection:

    def __init__(self, auth, base_url, skip_ssl_cert_check, reuse=True):
        self.auth = auth
        self.base_url = base_url
        self.skip_ssl_cert_check = skip_ssl_cert_check
        self._headers = {'User-Agent': 'qds-sdk-py-%s' % pkg_resources.get_distribution("qds-sdk").version,
                         'Content-Type': 'application/json'}

        self.reuse = reuse
        if reuse:
            self.session = requests.Session()
            self.session.mount('https://', MyAdapter())

    @retry((RetryWithDelay, requests.Timeout), tries=6, delay=30, backoff=2)
    def get_raw(self, path, params=None):
        return self._api_call_raw("GET", path, params=params)

    @retry((RetryWithDelay, requests.Timeout), tries=6, delay=30, backoff=2)
    def get(self, path, params=None):
        return self._api_call("GET", path, params=params)

    def put(self, path, data=None):
        return self._api_call("PUT", path, data)

    def post(self, path, data=None):
        return self._api_call("POST", path, data)

    def delete(self, path, data=None):
        return self._api_call("DELETE", path, data)

    def _api_call_raw(self, req_type, path, data=None, params=None):
        url = self.base_url.rstrip('/') + '/' + path

        if self.reuse:
            x = self.session
        else:
            x = requests

        kwargs = {'headers': self._headers, 'auth': self.auth, 'verify': not self.skip_ssl_cert_check}

        if data:
            kwargs['data'] = json.dumps(data)
        if params:
            kwargs['params'] = params

        log.info("[%s] %s" % (req_type, url))
        log.info("Payload: %s" % json.dumps(data, indent=4))
        log.info("Params: %s" % params)

        if req_type == 'GET':
            r = x.get(url, timeout=(30, 300), **kwargs)
        elif req_type == 'POST':
            r = x.post(url, timeout=(30, 300), **kwargs)
        elif req_type == 'PUT':
            r = x.put(url, timeout=(30, 300), **kwargs)
        elif req_type == 'DELETE':
            r = x.delete(url, timeout=(30, 300), **kwargs)
        else:
            raise NotImplemented

        self._handle_error(r)
        return r

    def _api_call(self, req_type, path, data=None, params=None):
        return self._api_call_raw(req_type, path, data=data, params=params).json()

    def _handle_error(self, response):
        """Raise exceptions in response to any http errors

        Args:
            response: A Response object

        Raises:
            BadRequest: if HTTP error code 400 returned.
            UnauthorizedAccess: if HTTP error code 401 returned.
            ForbiddenAccess: if HTTP error code 403 returned.
            ResourceNotFound: if HTTP error code 404 is returned.
            MethodNotAllowed: if HTTP error code 405 is returned.
            ResourceConflict: if HTTP error code 409 is returned.
            ResourceInvalid: if HTTP error code 422 is returned.
            ClientError: if HTTP error code falls in 401 - 499.
            ServerError: if HTTP error code falls in 500 - 599.
            ConnectionError: if unknown HTTP error code returned.
        """
        code = response.status_code

        if 200 <= code < 400:
            return

        if code == 400:
            sys.stderr.write(response.text + "\n")
            raise BadRequest(response)
        elif code == 401:
            sys.stderr.write(response.text + "\n")
            raise UnauthorizedAccess(response)
        elif code == 403:
            sys.stderr.write(response.text + "\n")
            raise ForbiddenAccess(response)
        elif code == 404:
            sys.stderr.write(response.text + "\n")
            raise ResourceNotFound(response)
        elif code == 405:
            sys.stderr.write(response.text + "\n")
            raise MethodNotAllowed(response)
        elif code == 409:
            sys.stderr.write(response.text + "\n")
            raise ResourceConflict(response)
        elif code == 422:
            sys.stderr.write(response.text + "\n")
            raise ResourceInvalid(response)
        elif code in (449, 503):
            sys.stderr.write(response.text + "\n")
            raise RetryWithDelay(response)
        elif 401 <= code < 500:
            sys.stderr.write(response.text + "\n")
            raise ClientError(response)
        elif 500 <= code < 600:
            sys.stderr.write(response.text + "\n")
            raise ServerError(response)
        else:
            raise ConnectionError(response)



"""
The actions module contains the definition for actions started by scheduler.
"""

from __future__ import print_function
import json

from qds_sdk.qubole import Qubole
from qds_sdk.resource import Resource
from argparse import ArgumentParser
from qds_sdk.commands import *

import logging

log = logging.getLogger("qds_actions")

class ActionCmdLine:
    """
    qds_sdk.ActionsCmdLine is the interface used by qds.py.
    """

    @staticmethod
    def parsers():
        argparser = ArgumentParser(prog="qds.py action",
                                        description="Scheduler action client for Qubole Data Service.")
        subparsers = argparser.add_subparsers()

        #List
        list = subparsers.add_parser("list",
                                     help="List actions for all schedulers")
        list.add_argument("--per-page", dest="per_page",
                          help="Number of items per page")
        list.add_argument("--page", dest="page",
                          help="Page Number")
        list.add_argument("--fields", nargs="*", dest="fields",
                          help="List of fields to show")
        list.set_defaults(func=ActionCmdLine.list)

        #View
        view = subparsers.add_parser("view",
                                     help="View a specific action")
        view.add_argument("id", help="Numeric id or name of the action")
        view.add_argument("--fields", nargs="*", dest="fields",
                          help="List of fields to show")
        view.set_defaults(func=ActionCmdLine.view)

        #Kill
        kill = subparsers.add_parser("kill",
                                     help="Kill a specific schedule action")
        kill.add_argument("id", help="Numeric id or name of the action")
        kill.set_defaults(func=ActionCmdLine.kill)

        #rerun
        rerun = subparsers.add_parser("rerun",
                                     help="rerun a specific schedule action")
        rerun.add_argument("id", help="Numeric id or name of the action")
        rerun.set_defaults(func=ActionCmdLine.rerun)

        #logs
        logs = subparsers.add_parser("logs",
                                     help="logs for a specific schedule action")
        logs.add_argument("id", help="Numeric id or name of the action")
        logs.set_defaults(func=ActionCmdLine.logs)

        #resultss
        results = subparsers.add_parser("results",
                                     help="results for a specific schedule action")
        results.add_argument("id", help="Numeric id or name of the action")
        results.set_defaults(func=ActionCmdLine.results)

        return argparser

    @staticmethod
    def run(args):
        parser = ActionCmdLine.parsers()
        parsed = parser.parse_args(args)
        return parsed.func(parsed)

    @staticmethod
    def filter_fields(act, fields):
        filtered = {}
        for field in fields:
            filtered[field] = act[field]
        return filtered

    @staticmethod
    def list(args):
        actionlist = Action.list(args.page, args.per_page)
        if args.fields:
            for a in actionlist:
                a.attributes = ActionCmdLine.filter_fields(a.attributes, args.fields)
        return json.dumps(actionlist, default=lambda o: o.attributes, sort_keys=True, indent=4)

    @staticmethod
    def view(args):
        act = Action.find(args.id)
        if args.fields:
            act.attributes = ActionCmdLine.filter_fields(act.attributes, args.fields)
        return json.dumps(act.attributes, sort_keys=True, indent=4)

    @staticmethod
    def kill(args):
        conn = Qubole.agent()
        ret_val = conn.put(Action.element_path(args.id) + "/kill", data=None)
        return json.dumps(ret_val, sort_keys=True, indent=4)

    @staticmethod
    def rerun(args):
        conn = Qubole.agent()
        ret_val = conn.post(Action.element_path(args.id) + "/rerun", data=None)
        return json.dumps(ret_val, sort_keys=True, indent=4)

    @staticmethod
    def logs(args):
        action = Action.find(args.id)
        print(action.logs())

    @staticmethod
    def results(args):
        action = Action.find(args.id)
        action.results()

class Action(Resource):

    """
    qds_sdk.Action is the Qubole action class.
    """

    """all actions use the /actions endpoint"""
    rest_entity_path = "actions"

    def getcommand(self):
        cmd = self.attributes["command"]
        if cmd is None:
            return None
        cmdclass = globals()[cmd["command_type"]]
        obj = cmdclass(cmd)
        return obj

    @staticmethod
    def list(page = None, per_page = None):
        conn = Qubole.agent()
        url_path = Action.rest_entity_path
        params = {}
        if page is not None:
            params['page'] = page
        if per_page is not None:
            params['per_page'] = per_page

        #Todo Page numbers are thrown away right now
        actjson = conn.get(url_path, params)
        actlist = []
        for a in actjson["actions"]:
            actlist.append(Action(a))
        return actlist

    def kill(self):
        conn = Qubole.agent()
        return conn.put(self.element_path(self.id) + "/kill", data=None)

    def rerun(self):
        conn = Qubole.agent()
        return conn.post(self.element_path(self.id) + "/rerun", data=None)

    def status(self):
        return self.attributes["status"]

    def logs(self):
        if self.status().lower() == "submitted":
            return "Logs for action are not yet available. Action is waiting for underlying command to be created."
        if self.status().lower() == "waiting":
            return "Logs for action are not yet available. Waiting for dependencies to be met."
        if self.status().lower() == "not_found":
            return "Logs for action are not available. Dependencies not found."
        if self.status().lower() == "cancelled":
            return "Logs for action are not available. Action was cancelled before underlying command gets created."
        cmd = self.getcommand()
        if cmd is None:
            return "Logs for action are not yet available."
        else:
            return cmd.get_log()

    def results(self):
        cmd = self.getcommand()
        if cmd is None:
            print("Results for action are not yet available.")
        else:
            cmd.get_results()





import json

from qds_sdk.qubole import Qubole
from qds_sdk.resource import Resource
from argparse import ArgumentParser


class RoleCmdLine:
    """
    qds_sdk.RoleCmdLine is the interface used by qds.py.
    """

    @staticmethod
    def parsers():
        argparser = ArgumentParser(prog="qds.py role",
                                   description="Client to Manage Roles in Qubole Data Service.")
        subparsers = argparser.add_subparsers()

        # Create
        create = subparsers.add_parser("create", help="Create a new Role")
        create.add_argument("--name", dest="name", required=True, help="Name of the new Role")
        create.add_argument("--policy", dest="policy", required=True,
                            help="Policy Statement example '[{\"access\":\"deny\", \"resource\": \"all\", \"action\": \"[\"create\",\"update\",\"delete\"\]\"}]'")
        create.set_defaults(func=RoleCmdLine.create)

        #List
        list = subparsers.add_parser("list", help="List all Roles")
        list.add_argument("--per-page", dest="per_page", help="Number of items per page")
        list.add_argument("--page", dest="page", help="Page Number")
        list.set_defaults(func=RoleCmdLine.list)

        #View
        view = subparsers.add_parser("view", help="View a specific Role")
        view.add_argument("id", help="Numeric id of the Role")
        view.set_defaults(func=RoleCmdLine.view)

        #update
        update = subparsers.add_parser("update", help="Update a specific Role's name or policy")
        update.add_argument("id", help="Numeric id of the Role")
        update.add_argument("--name", dest="name", help="New name of the Role")
        update.add_argument("--policy", dest="policy",
                            help="Policy Statement example '[{\"access\":\"deny\", \"resource\": \"all\", \"action\": \"[\"create\",\"update\",\"delete\"\]\"}]'")
        update.set_defaults(func=RoleCmdLine.update)

        #Delete
        delete = subparsers.add_parser("delete", help="Delete a Role")
        delete.add_argument("id", help="Numeric id of the Role to be deleted")
        delete.set_defaults(func=RoleCmdLine.delete)

        #duplicate
        duplicate = subparsers.add_parser("duplicate", help="Duplicates/Clones a Role")
        duplicate.add_argument("id", help="Numeric id of the Role to be cloned")
        duplicate.add_argument("--name", dest="name", required=False, help="Name of the new Role")
        duplicate.add_argument("--policy", dest="policy", required=False,
                               help="Optional policy Statement example '[{\"access\":\"allow\", \"resource\": \"all\"}]'")
        duplicate.set_defaults(func=RoleCmdLine.duplicate)

        #Assign Role
        assign_role = subparsers.add_parser("assign-role", help="Assigns a Role to a Group")
        assign_role.add_argument("id", help="Numeric id of the Role")
        assign_role.add_argument("--group-id", dest="group_id", required=True, help="Numeric Id of the Group")
        assign_role.set_defaults(func=RoleCmdLine.assign_role)

        #UnAssign Role
        unassign_role = subparsers.add_parser("unassign-role", help="Unassigna a Role from a Group")
        unassign_role.add_argument("id", help="Numeric id of the Role")
        unassign_role.add_argument("--group-id", dest="group_id", required=True, help="Numeric Id of the Group")
        unassign_role.set_defaults(func=RoleCmdLine.unassign_role)

        #List groups
        list_groups = subparsers.add_parser("list-groups", help="List all Groups for a Role ")
        list_groups.add_argument("id", help="Numeric id of the Role")
        list_groups.set_defaults(func=RoleCmdLine.list_groups)

        return argparser

    @staticmethod
    def run(args):
        parser = RoleCmdLine.parsers()
        parsed = parser.parse_args(args)
        return parsed.func(parsed)

    @staticmethod
    def create(args):
        role = Role.create(name=args.name, policies=args.policy)
        return json.dumps(role.attributes, sort_keys=True, indent=4)

    @staticmethod
    def list(args):
        rolelist = Role.list(args.page, args.per_page)
        return json.dumps(rolelist, default=lambda o: o.attributes, sort_keys=True, indent=4)

    @staticmethod
    def view(args):
        role = Role.find(args.id)
        return json.dumps(role.attributes, sort_keys=True, indent=4)

    @staticmethod
    def update(args):
        options = {}
        if args.name is not None:
            options["name"] = args.name
        if args.policy is not None:
            options["policies"] = args.policy
        return json.dumps(Role.update(args.id, **options), sort_keys=True, indent=4)

    @staticmethod
    def delete(args):
        return json.dumps(Role.delete(args.id), sort_keys=True, indent=4)

    @staticmethod
    def duplicate(args):
        options = {}
        if args.name is not None:
            options["name"] = args.name
        if args.policy is not None:
            options["policy"] = args.policy
        return json.dumps(Role.duplicate(args.id, **options), sort_keys=True, indent=4)

    @staticmethod
    def assign_role(args):
        return Role.assign_role(args.id, args.group_id)

    @staticmethod
    def unassign_role(args):
        return Role.unassign_role(args.id, args.group_id)

    @staticmethod
    def list_groups(args):
        return json.dumps(Role.list_groups(args.id), sort_keys=True, indent=4)


class Role(Resource):
    """
    qds_sdk.Role is the base Qubole Role class.
    """

    """ all commands use the /role endpoint"""

    rest_entity_path = "roles"

    @staticmethod
    def list(page=None, per_page=None):
        conn = Qubole.agent()
        url_path = Role.rest_entity_path
        page_attr = []
        if page is not None:
            page_attr.append("page=%s" % page)
        if per_page is not None:
            page_attr.append("per_page=%s" % per_page)
        if page_attr:
            url_path = "%s?%s" % (Role.rest_entity_path, "&".join(page_attr))

        rolejson = conn.get(url_path)
        rolelist = []
        for s in rolejson["roles"]:
            rolelist.append(Role(s))
        return rolelist

    @staticmethod
    def update(role_id, **kwargs):
        conn = Qubole.agent()
        url_path = "roles/%s" % role_id
        return conn.put(url_path, data=kwargs)

    @staticmethod
    def delete(role_id):
        conn = Qubole.agent()
        url_path = "roles/%s" % role_id
        return conn.delete(url_path)

    @staticmethod
    def duplicate(role_id, **kwargs):
        conn = Qubole.agent()
        url_path = "roles/%s/duplicate" % role_id
        return conn.post(url_path, data=kwargs)

    @staticmethod
    def assign_role(role_id, qbol_group_id):
        conn = Qubole.agent()
        url_path = "groups/%s/roles/%s/assign" % (qbol_group_id, role_id)
        return conn.put(url_path)

    @staticmethod
    def unassign_role(role_id, qbol_group_id):
        conn = Qubole.agent()
        url_path = "groups/%s/roles/%s/unassign" % (qbol_group_id, role_id)
        return conn.put(url_path)

    @staticmethod
    def list_groups(role_id):
        conn = Qubole.agent()
        url_path = "roles/%s/groups" % role_id
        return conn.get(url_path)





#!/bin/env python

from __future__ import print_function
from qds_sdk.qubole import Qubole
from qds_sdk.commands import *
from qds_sdk.cluster import *
import qds_sdk.exception
from qds_sdk.scheduler import SchedulerCmdLine
from qds_sdk.actions import ActionCmdLine
from qds_sdk.report import ReportCmdLine
from qds_sdk.dbtaps import DbTapCmdLine
from qds_sdk.role import RoleCmdLine
from qds_sdk.group import GroupCmdLine

import os
import sys
import traceback
import logging
import json
from optparse import OptionParser

log = logging.getLogger("qds")
CommandClasses = {
    "hivecmd": HiveCommand,
    "sparkcmd": SparkCommand,
    "dbtapquerycmd": DbTapQueryCommand,
    "pigcmd":  PigCommand,
    "hadoopcmd": HadoopCommand,
    "shellcmd": ShellCommand,
    "dbexportcmd": DbExportCommand,
    "dbimportcmd": DbImportCommand,
    "prestocmd": PrestoCommand
}

usage_str = ("Usage: \n"
             "qds [options] <CmdArgs|ClusterArgs|ReportArgs>\n"
             "\nCmdArgs:\n" +
             "  <hivecmd|hadoopcmd|prestocmd|pigcmd|shellcmd|dbexportcmd|dbtapquerycmd|sparkcmd> <submit|run|check|cancel|getresult|getlog> [args .. ]\n"
             "  submit [cmd-specific-args .. ] : submit cmd & print id \n"
             "  run [cmd-specific-args .. ] : submit cmd & wait. print results \n"
             "  check <id> : print the cmd object for this Id\n"
             "  cancel <id> : cancels the cmd with this Id\n"
             "  getresult <id> : get the results for the cmd with this Id\n"
             "  getlog <id> : get the logs for the cmd with this Id\n"
             "\nClusterArgs:\n" +
             "  cluster <create|delete|update|list|start|terminate|status|reassign_label> [args .. ]\n"
             "  create [cmd-specific-args ..] : create a new cluster\n"
             "  delete [cmd-specific-args ..] : delete an existing cluster\n"
             "  update [cmd-specific-args ..] : update the settings of an existing cluster\n"
             "  clone [cmd-specific-args ..] : clone a cluster from an existing one\n"
             "  list [cmd-specific-args ..] : list existing cluster(s)\n"
             "  start [cmd-specific-args ..] : start an existing cluster\n"
             "  terminate [cmd-specific-args ..] : terminate a running cluster\n"
             "  status [cmd-specific-args ..] : show whether the cluster is up or down\n" +
             "  reassign_label [cmd-specific-args ..] : reassign label from one cluster to another\n" +
             "\nDbTap:\n" +
             "  dbtap --help\n" +
             "\nReportArgs:\n" +
             "  report (<report-name> [options] | list)\n" +
             "\nGroup:\n" +
             "  group --help\n" +
             "\nRole:\n" +
             "  role --help\n" +
            "\nScheduler:\n" +
             "  scheduler --help\n")


def usage(parser=None):
    if parser is None:
        sys.stderr.write(usage_str)
    else:
        parser.print_help()
    sys.exit(1)


def checkargs_id(args):
    if len(args) != 1:
        sys.stderr.write("expecting single argument command id\n")
        usage()


def submitaction(cmdclass, args):
    args = cmdclass.parse(args)
    if args is not None:
        args.pop("print_logs") # This is only useful while using the 'run' action.
        cmd = cmdclass.create(**args)
        print("Submitted %s, Id: %s" % (cmdclass.__name__, cmd.id))
        return 0


def _getresult(cmdclass, cmd):
    if Command.is_success(cmd.status):
        log.info("Fetching results for %s, Id: %s" % (cmdclass.__name__, cmd.id))
        cmd.get_results(sys.stdout, delim='\t')
        return 0
    else:
        log.error("Cannot fetch results - command Id: %s failed with status: %s" % (cmd.id, cmd.status))
        return 1


def runaction(cmdclass, args):
    args = cmdclass.parse(args)
    if args is not None:
        print_logs = args.pop("print_logs") # We don't want to send this to the API.
        cmd = cmdclass.run(**args)
        if print_logs:
            sys.stderr.write(cmd.get_log())
        return _getresult(cmdclass, cmd)


def checkaction(cmdclass, args):
    checkargs_id(args)
    o = cmdclass.find(args.pop(0))
    print(str(o))
    return 0


def cancelaction(cmdclass, args):
    checkargs_id(args)
    r = cmdclass.cancel_id(args.pop(0))
    skey = 'kill_succeeded'
    if r.get(skey) is None:
        sys.stderr.write("Invalid Json Response %s - missing field '%s'" % (str(r), skey))
        return 11
    elif r['kill_succeeded']:
        print("Command killed successfully")
        return 0
    else:
        sys.stderr.write("Cancel failed with reason '%s'\n" % r.get('result'))
        return 12


def getresultaction(cmdclass, args):
    checkargs_id(args)
    cmd = cmdclass.find(args.pop(0))
    return _getresult(cmdclass, cmd)


def getlogaction(cmdclass, args):
    checkargs_id(args)
    print(cmdclass.get_log_id(args.pop(0)))
    return 0


def getjobsaction(cmdclass, args):
    checkargs_id(args)
    cmd = cmdclass.find(args.pop(0))
    if Command.is_done(cmd.status):
        log.info("Fetching jobs for %s, Id: %s" % (cmdclass.__name__, cmd.id))
        print(cmdclass.get_jobs_id(cmd.id))
        return 0
    else:
        log.error("Cannot fetch jobs - command Id: %s is not done. Status: %s" % (cmd.id, cmd.status))
        return 1


def cmdmain(cmd, args):
    cmdclass = CommandClasses[cmd]

    actionset = set(["submit", "run", "check", "cancel", "getresult", "getlog", "getjobs"])
    if len(args) < 1:
        sys.stderr.write("missing argument containing action\n")
        usage()

    action = args.pop(0)
    if action not in actionset:
        sys.stderr.write("action must be one of <%s>\n" % "|".join(actionset))
        usage()

    return globals()[action + "action"](cmdclass, args)


def checkargs_cluster_id_label(args):
    if len(args) != 1:
        sys.stderr.write("expecting single argument cluster id or cluster label\n")
        usage()


def cluster_create_action(clusterclass, args):
    arguments = clusterclass._parse_create_update(args, action="create")
    cluster_info = _create_cluster_info(arguments)
    result = clusterclass.create(cluster_info.minimal_payload())
    print(json.dumps(result, indent=4))
    return 0


def cluster_update_action(clusterclass, args):
    arguments = clusterclass._parse_create_update(args, action="update")
    cluster_info = _create_cluster_info(arguments)
    result = clusterclass.update(arguments.cluster_id_label, cluster_info.minimal_payload())
    print(json.dumps(result, indent=4))
    return 0

def cluster_clone_action(clusterclass, args):
    arguments = clusterclass._parse_create_update(args, action="clone")
    cluster_info = _create_cluster_info(arguments)
    result = clusterclass.clone(arguments.cluster_id_label, cluster_info.minimal_payload())
    print(json.dumps(result, indent=4))
    return 0

def _create_cluster_info(arguments):
    cluster_info = ClusterInfo(arguments.label,
                               arguments.aws_access_key_id,
                               arguments.aws_secret_access_key,
                               arguments.disallow_cluster_termination,
                               arguments.enable_ganglia_monitoring,
                               arguments.node_bootstrap_file,)

    cluster_info.set_ec2_settings(arguments.aws_region,
                                  arguments.aws_availability_zone,
                                  arguments.vpc_id,
                                  arguments.subnet_id)

    custom_config = None
    if arguments.custom_config_file is not None:
        try:
            custom_config = open(arguments.custom_config_file).read()
        except IOError as e:
            sys.stderr.write("Unable to read custom config file: %s\n" %
                             str(e))
            usage()

    cluster_info.set_hadoop_settings(arguments.master_instance_type,
                                     arguments.slave_instance_type,
                                     arguments.initial_nodes,
                                     arguments.max_nodes,
                                     custom_config,
                                     arguments.slave_request_type,
                                     arguments.use_hbase,
                                     arguments.custom_ec2_tags,
                                     arguments.use_hadoop2)

    cluster_info.set_spot_instance_settings(
          arguments.maximum_bid_price_percentage,
          arguments.timeout_for_request,
          arguments.maximum_spot_instance_percentage)

    cluster_info.set_stable_spot_instance_settings(
          arguments.stable_maximum_bid_price_percentage,
          arguments.stable_timeout_for_request,
          arguments.stable_allow_fallback)

    fairscheduler_config_xml = None
    if arguments.fairscheduler_config_xml_file is not None:
        try:
            fairscheduler_config_xml = open(arguments.fairscheduler_config_xml_file).read()
        except IOError as e:
            sys.stderr.write("Unable to read config xml file: %s\n" %
                             str(e))
            usage()
    cluster_info.set_fairscheduler_settings(fairscheduler_config_xml,
                                            arguments.default_pool)

    customer_ssh_key = None
    if arguments.customer_ssh_key_file is not None:
        try:
            customer_ssh_key = open(arguments.customer_ssh_key_file).read()
        except IOError as e:
            sys.stderr.write("Unable to read customer ssh key file: %s\n" %
                             str(e))
            usage()
    cluster_info.set_security_settings(arguments.encrypted_ephemerals,
                                       customer_ssh_key)

    presto_custom_config = None
    if arguments.presto_custom_config_file is not None:
        try:
            presto_custom_config = open(arguments.presto_custom_config_file).read()
        except IOError as e:
            sys.stderr.write("Unable to read presto custom config file: %s\n" %
                             str(e))
            usage()
    cluster_info.set_presto_settings(arguments.enable_presto,
                                     presto_custom_config)

    return cluster_info


def cluster_delete_action(clusterclass, args):
    checkargs_cluster_id_label(args)
    result = clusterclass.delete(args.pop(0))
    print(json.dumps(result, indent=4))
    return 0


def cluster_list_action(clusterclass, args):
    arguments = clusterclass._parse_list(args)
    if arguments['cluster_id'] is not None:
        result = clusterclass.show(arguments['cluster_id'])
    elif arguments['label'] is not None:
        result = clusterclass.show(arguments['label'])
    elif arguments['state'] is not None:
        result = clusterclass.list(state=arguments['state'])
    else:
        result = clusterclass.list()
    print(json.dumps(result, indent=4))
    return 0


def cluster_start_action(clusterclass, args):
    checkargs_cluster_id_label(args)
    result = clusterclass.start(args.pop(0))
    print(json.dumps(result, indent=4))
    return 0


def cluster_terminate_action(clusterclass, args):
    checkargs_cluster_id_label(args)
    result = clusterclass.terminate(args.pop(0))
    print(json.dumps(result, indent=4))
    return 0


def cluster_status_action(clusterclass, args):
    checkargs_cluster_id_label(args)
    result = clusterclass.status(args.pop(0))
    print(json.dumps(result, indent=4))
    return 0


def cluster_reassign_label_action(clusterclass, args):
    arguments = clusterclass._parse_reassign_label(args)
    result = clusterclass.reassign_label(arguments.destination_cluster,
                                         arguments.label)
    print(json.dumps(result, indent=4))
    return 0


def clustermain(args):
    clusterclass = Cluster
    actionset = set(["create", "delete", "update", "clone", "list", "start", "terminate", "status", "reassign_label"])

    if len(args) < 1:
        sys.stderr.write("missing argument containing action\n")
        usage()

    action = args.pop(0)
    if action not in actionset:
        sys.stderr.write("action must be one of <%s>\n" % "|".join(actionset))
        usage()
    return globals()["cluster_" + action + "_action"](clusterclass, args)


def reportmain(args):
    result = ReportCmdLine.run(args)
    print(result)


def actionmain(args):
    result = ActionCmdLine.run(args)
    print(result)

def schedulermain(args):
    result = SchedulerCmdLine.run(args)
    print(result)

def dbtapmain(args):
    result = DbTapCmdLine.run(args)
    print(result)

def rolemain(args):
    result = RoleCmdLine.run(args)
    print(result)

def groupmain(args):
    result = GroupCmdLine.run(args)
    print(result)

def main():

    optparser = OptionParser(usage=usage_str)
    optparser.add_option("--token", dest="api_token",
                         default=os.getenv('QDS_API_TOKEN'),
                         help="api token for accessing Qubole. must be specified via command line or passed in via environment variable QDS_API_TOKEN")

    optparser.add_option("--url", dest="api_url",
                         default=os.getenv('QDS_API_URL'),
                         help="base url for QDS REST API. defaults to https://api.qubole.com/api ")

    optparser.add_option("--version", dest="api_version",
                         default=os.getenv('QDS_API_VERSION'),
                         help="version of REST API to access. defaults to v1.2")

    optparser.add_option("--poll_interval", dest="poll_interval",
                         default=os.getenv('QDS_POLL_INTERVAL'),
                         help="interval for polling API for completion and other events. defaults to 5s")

    optparser.add_option("--skip_ssl_cert_check", dest="skip_ssl_cert_check", action="store_true",
                         default=False,
                         help="skip verification of server SSL certificate. Insecure: use with caution.")

    optparser.add_option("-v", dest="verbose", action="store_true",
                         default=False,
                         help="verbose mode - info level logging")

    optparser.add_option("--vv", dest="chatty", action="store_true",
                         default=False,
                         help="very verbose mode - debug level logging")

    optparser.disable_interspersed_args()
    (options, args) = optparser.parse_args()

    if options.chatty:
        logging.basicConfig(level=logging.DEBUG)
    elif options.verbose:
        logging.basicConfig(level=logging.INFO)
    else:
        logging.basicConfig(level=logging.WARN)

    if options.api_token is None:
        sys.stderr.write("No API Token provided\n")
        usage(optparser)

    if options.api_url is None:
        options.api_url = "https://api.qubole.com/api/"

    if options.api_version is None:
        options.api_version = "v1.2"

    if options.poll_interval is None:
        options.poll_interval = 5

    if options.skip_ssl_cert_check is None:
        options.skip_ssl_cert_check = False
    elif options.skip_ssl_cert_check:
        log.warn("Insecure mode enabled: skipping SSL cert verification\n")

    Qubole.configure(api_token=options.api_token,
                     api_url=options.api_url,
                     version=options.api_version,
                     poll_interval=options.poll_interval,
                     skip_ssl_cert_check=options.skip_ssl_cert_check)

    if len(args) < 1:
        sys.stderr.write("Missing first argument containing command type\n")
        usage(optparser)

    a0 = args.pop(0)
    if a0 in CommandClasses:
        return cmdmain(a0, args)

    if a0 == "cluster":
        return clustermain(args)

    if a0 == "action":
        return actionmain(args)

    if a0 == "scheduler":
        return schedulermain(args)

    if a0 == "report":
        return reportmain(args)

    if a0 == "dbtap":
        return dbtapmain(args)

    if a0 == "group":
        return groupmain(args)

    if a0 == "role":
        return rolemain(args)

    cmdset = set(CommandClasses.keys())
    sys.stderr.write("First command must be one of <%s>\n" %
                     "|".join(cmdset.union(["cluster", "scheduler", "report",
                       "dbtap", "role", "group"])))
    usage(optparser)


if __name__ == '__main__':
    try:
        sys.exit(main())
    except qds_sdk.exception.Error as e:
        sys.stderr.write("Error: Status code %s (%s) from url %s\n" %
                         (e.request.status_code, e.__class__.__name__,
                          e.request.url))
        sys.exit(1)
    except qds_sdk.exception.ParseError as e:
        sys.stderr.write("Error: %s\n" % str(e))
        sys.stderr.write("Usage: %s\n" % e.usage)
        sys.exit(2)
    except Exception:
        traceback.print_exc(file=sys.stderr)
        sys.exit(3)



import requests
from qds_sdk.connection import Connection
from qds_sdk.exception import ConfigError


class QuboleAuth(requests.auth.AuthBase):
    def __init__(self, token):
        self.api_token = token

    def __call__(self, r):
        r.headers['X-AUTH-TOKEN'] = self.api_token
        return r


class Qubole:
    """
    Singleton for storing authorization credentials and other
    configuration parameters for QDS.
    """

    _auth = None
    api_token = None
    base_url = None
    poll_interval = None
    skip_ssl_cert_check = None

    @classmethod
    def configure(cls, api_token,
                  api_url="https://api.qubole.com/api/", version="v1.2",
                  poll_interval=5, skip_ssl_cert_check=False):
        """
        Set parameters governing interaction with QDS

        Args:
            `api_token`: authorization token for QDS. required

            `api_url`: the base URL for QDS API. configurable for testing only

            `version`: QDS REST api version

            `poll_interval`: interval in secs when polling QDS for events
        """
        cls._auth = QuboleAuth(api_token)
        cls.api_token = api_token
        cls.base_url = api_url.rstrip('/') + '/' + version
        cls.poll_interval = poll_interval
        cls.skip_ssl_cert_check = skip_ssl_cert_check

    cached_agent = None

    @classmethod
    def agent(cls):
        """
        Returns:
           a connection object to make REST calls to QDS
        """
        if cls.api_token is None:
            raise ConfigError("No API Token specified - please supply one via Qubole.configure()")

        if cls.cached_agent is None:
            cls.cached_agent = Connection(cls._auth, cls.base_url, cls.skip_ssl_cert_check)

        return cls.cached_agent



"""
The report module contains the definitions for retrieving various reports.
"""
from qds_sdk.qubole import Qubole
from qds_sdk.resource import Resource
import argparse
import json


class ReportCmdLine:
    """
    qds_sdk.ReportCmdLine is the interface used by qds.py.
    """

    @staticmethod
    def parsers():
        argparser = argparse.ArgumentParser(prog="qds.py report",
                description="Report client for Qubole Data Service.")
        subparsers = argparser.add_subparsers(title="report name")

        # For listing the names of all reports
        index = subparsers.add_parser("list",
                description="show names of all available reports")
        index.set_defaults(func=ReportCmdLine.index)

        # Canonical Hive Commands Report
        chc = subparsers.add_parser("canonical_hive_commands",
                description="Show report for canonical hive commands")
        chc.add_argument("--start-date", default=argparse.SUPPRESS,
                help="""The date from which you want the report.  The report
                may contain some data prior to this date also.  api default =
                beginning of time""")
        chc.add_argument("--end-date", default=argparse.SUPPRESS,
                help="""The date till which you want the report.  The report
                may contain some date after this date also.  api default =
                now""")
        chc.add_argument("--offset", type=int, default=argparse.SUPPRESS,
                help="""The starting point of the results.  api default = 0""")
        chc.add_argument("--limit", type=int, default=argparse.SUPPRESS,
                help="""The number of results to fetch.  api default = 10""")
        chc.add_argument("--sort", dest="sort_column",
                default=argparse.SUPPRESS, choices=["frequency", "cpu",
                "fs_bytes_read", "fs_bytes_written"], help="""The column used
                to sort the report. Since this report returns top canonical
                hive commands, the sort order is always descending.  api
                default = frequency""")
        chc.add_argument("--show-ast", default=argparse.SUPPRESS,
                action="store_true", help="""Also return the serialized AST
                corresponding to the canonical query. By default, only the
                canonical_query_id is return.""")
        chc.set_defaults(func=ReportCmdLine.canonical_hive_commands)


        # All Commands Report
        ac = subparsers.add_parser("all_commands",
                description="Show report for all commands")
        ac.add_argument("--start-date", default=argparse.SUPPRESS,
                help="""The date from which you want the report (inclusive)
                api default = 7 days before the end date.""")
        ac.add_argument("--end-date", default=argparse.SUPPRESS,
                help="""The date till which you want the report (exclusive)
                api default = today""")
        ac.add_argument("--offset", type=int, default=argparse.SUPPRESS,
                help="""The starting point of the results.  api default = 0""")
        ac.add_argument("--limit", type=int, default=argparse.SUPPRESS,
                help="""The number of results to fetch.  api default = 10""")
        ac.add_argument("--sort", dest="sort_column",
                default=argparse.SUPPRESS, choices=["cpu", "fs_bytes_read",
                "fs_bytes_written", "time"], help="""The column used to sort the
                report.  api default = time (chronological order)""")
        ac.add_argument("--by-user", default=argparse.SUPPRESS,
                action="store_true", help="""Report only those queries which
                are created by the current user. By default, all queries from
                the current account are reported.""")
        ac.set_defaults(func=ReportCmdLine.all_commands)


        # Foo Bar Report
        #fb = subparsers.add_parser("foo_bar",
        #        description="Show report for foo bar")
        #fb.add_argument("--start-date", default=argparse.SUPPRESS,
        #        help="""The date from which you want the report.
        #        api default = """)
        #fb.add_argument("--end-date", default=argparse.SUPPRESS,
        #        help="""The date till which you want the report.
        #        api default = """)
        #fb.add_argument("--offset", type=int, default=argparse.SUPPRESS,
        #        help="""The starting point of the results.  api default = """)
        #fb.add_argument("--limit", type=int, default=argparse.SUPPRESS,
        #        help="""The number of results to fetch.  api default = """)
        #fb.add_argument("--sort", dest="sort_column",
        #        default=argparse.SUPPRESS, help="""The column used to sort the
        #        report.  api default = """)
        #fb.add_argument("--sort-order", default=argparse.SUPPRESS,
        #        help="""The sorting order used.  api default = """)
        #fb.add_argument("--any-other", default=argparse.SUPPRESS,
        #        help="""Any other argument specific to this report.
        #        api default = """)
        #fb.set_defaults(func=ReportCmdLine.foo_bar)

        return argparser

    @staticmethod
    def run(args):
        parser = ReportCmdLine.parsers()
        parsed = parser.parse_args(args)
        return parsed.func(parsed)

    @staticmethod
    def index(args):
        result = Report.index()
        return json.dumps(result, indent=4)

    @staticmethod
    def canonical_hive_commands(args):
        data = vars(args)
        data.pop("func")    # We don't want to send this to the api
        result = Report.show("canonical_hive_commands", data)
        return json.dumps(result, indent=4)

    @staticmethod
    def all_commands(args):
        data = vars(args)
        data.pop("func")    # We don't want to send this to the api
        result = Report.show("all_commands", data)
        return json.dumps(result, indent=4)

    #@staticmethod
    #def foo_bar(args):
    #    data = vars(args)
    #    data.pop("func")    # We don't want to send this to the api
    #    result = Report.show("foo_bar", data)
    #    return json.dumps(result, indent=4)


class Report(Resource):
    """
    qds_sdk.Report is the base Qubole Reports class.
    """

    """all reports use the /reports endpoint"""
    rest_entity_path = "reports"

    @classmethod
    def show(cls, report_name, data):
        """
        Shows a report by issuing a GET request to the /reports/report_name
        endpoint.

        Args:
            `report_name`: the name of the report to show

            `data`: the parameters for the report
        """
        conn = Qubole.agent()
        return conn.get(cls.element_path(report_name), data)

    @classmethod
    def index(cls):
        """
        Shows a list of all available reports by issuing a GET request to the
        /reports endpoint.
        """
        conn = Qubole.agent()
        return conn.get(cls.rest_entity_path)



"""
The cluster module contains the definitions for retrieving and manipulating
cluster information.
"""

from qds_sdk.qubole import Qubole
from qds_sdk.resource import Resource
from argparse import ArgumentParser

import logging
import json

log = logging.getLogger("qds_cluster")


def str2bool(v):
    return v.lower() in ("yes", "true", "t", "1")


class Cluster(Resource):
    """
    qds_sdk.Cluster is the class for retrieving and manipulating cluster
    information.
    """

    rest_entity_path = "clusters"

    @classmethod
    def _parse_list(cls, args):
        """
        Parse command line arguments to construct a dictionary of cluster
        parameters that can be used to determine which clusters to list.

        Args:
            `args`: sequence of arguments

        Returns:
            Dictionary that can be used to determine which clusters to list
        """
        argparser = ArgumentParser(prog="cluster list")

        group = argparser.add_mutually_exclusive_group()

        group.add_argument("--id", dest="cluster_id",
                           help="show cluster with this id")

        group.add_argument("--label", dest="label",
                           help="show cluster with this label")

        group.add_argument("--state", dest="state", action="store",
                           choices=['up', 'down', 'pending', 'terminating'],
                           help="list only clusters in the given state")

        arguments = argparser.parse_args(args)
        return vars(arguments)

    @classmethod
    def list(cls, state=None):
        """
        List existing clusters present in your account.

        Kwargs:
            `state`: list only those clusters which are in this state

        Returns:
            List of clusters satisfying the given criteria
        """
        conn = Qubole.agent()
        if state is None:
            return conn.get(cls.rest_entity_path)
        elif state is not None:
            cluster_list = conn.get(cls.rest_entity_path)
            result = []
            for cluster in cluster_list:
                if state.lower() == cluster['cluster']['state'].lower():
                    result.append(cluster)
            return result

    @classmethod
    def show(cls, cluster_id_label):
        """
        Show information about the cluster with id/label `cluster_id_label`.
        """
        conn = Qubole.agent()
        return conn.get(cls.element_path(cluster_id_label))

    @classmethod
    def status(cls, cluster_id_label):
        """
        Show the status of the cluster with id/label `cluster_id_label`.
        """
        conn = Qubole.agent()
        return conn.get(cls.element_path(cluster_id_label) + "/state")

    @classmethod
    def start(cls, cluster_id_label):
        """
        Start the cluster with id/label `cluster_id_label`.
        """
        conn = Qubole.agent()
        data = {"state": "start"}
        return conn.put(cls.element_path(cluster_id_label) + "/state", data)

    @classmethod
    def terminate(cls, cluster_id_label):
        """
        Terminate the cluster with id/label `cluster_id_label`.
        """
        conn = Qubole.agent()
        data = {"state": "terminate"}
        return conn.put(cls.element_path(cluster_id_label) + "/state", data)

    @classmethod
    def _parse_create_update(cls, args, action):
        """
        Parse command line arguments to determine cluster parameters that can
        be used to create or update a cluster.

        Args:
            `args`: sequence of arguments

            `action`: "create", "update" or "clone"

        Returns:
            Object that contains cluster parameters
        """
        argparser = ArgumentParser(prog="cluster %s" % action)

        create_required = False
        label_required = False

        if action == "create":
            create_required = True
        elif action == "update":
            argparser.add_argument("cluster_id_label",
                                   help="id/label of the cluster to update")
        elif action == "clone":
            argparser.add_argument("cluster_id_label",
                                   help="id/label of the cluster to update")
            label_required = True

        argparser.add_argument("--label", dest="label",
                               nargs="+", required=(create_required or label_required),
                               help="list of labels for the cluster" +
                                    " (atleast one label is required)")

        ec2_group = argparser.add_argument_group("ec2 settings")
        ec2_group.add_argument("--access-key-id",
                               dest="aws_access_key_id",
                               required=create_required,
                               help="access key id for customer's aws" +
                                    " account. This is required while" +
                                    " creating the cluster",)
        ec2_group.add_argument("--secret-access-key",
                               dest="aws_secret_access_key",
                               required=create_required,
                               help="secret access key for customer's aws" +
                                    " account. This is required while" +
                                    " creating the cluster",)
        ec2_group.add_argument("--aws-region",
                               dest="aws_region",
                               choices=["us-east-1", "us-west-2", "ap-northeast-1",
                                        "eu-west-1", "ap-southeast-1", "us-west-1"],
                               help="aws region to create the cluster in",)
        ec2_group.add_argument("--aws-availability-zone",
                               dest="aws_availability_zone",
                               help="availability zone to" +
                                    " create the cluster in",)
        ec2_group.add_argument("--subnet-id",
                               dest="subnet_id",
                               help="subnet to create the cluster in",)
        ec2_group.add_argument("--vpc-id",
                               dest="vpc_id",
                               help="vpc to create the cluster in",)

        hadoop_group = argparser.add_argument_group("hadoop settings")
        hadoop_group.add_argument("--master-instance-type",
                                  dest="master_instance_type",
                                  help="instance type to use for the hadoop" +
                                       " master node",)
        hadoop_group.add_argument("--slave-instance-type",
                                  dest="slave_instance_type",
                                  help="instance type to use for the hadoop" +
                                       " slave nodes",)
        hadoop_group.add_argument("--initial-nodes",
                                  dest="initial_nodes",
                                  type=int,
                                  help="number of nodes to start the" +
                                       " cluster with",)
        hadoop_group.add_argument("--max-nodes",
                                  dest="max_nodes",
                                  type=int,
                                  help="maximum number of nodes the cluster" +
                                       " may be auto-scaled up to")
        hadoop_group.add_argument("--custom-config",
                                  dest="custom_config_file",
                                  help="location of file containg custom" +
                                       " hadoop configuration overrides")
        hadoop_group.add_argument("--slave-request-type",
                                  dest="slave_request_type",
                                  choices=["ondemand", "spot", "hybrid"],
                                  help="purchasing option for slave instaces",)
        hadoop_group.add_argument("--use-hbase", dest="use_hbase",
                                  action="store_true", default=None,
                                  help="Use hbase on this cluster",)

        hadoop2 = hadoop_group.add_mutually_exclusive_group()
        hadoop2.add_argument("--use-hadoop2",
                             dest="use_hadoop2",
                             action="store_true",
                             default=None,
                             help="Use hadoop2 instead of hadoop1")
        hadoop2.add_argument("--use-hadoop1",
                             dest="use_hadoop2",
                             action="store_false",
                             default=None,
                             help="Use hadoop1 instead of hadoop2. This is the default.")

        spot_group = argparser.add_argument_group("spot instance settings" +
                    " (valid only when slave-request-type is hybrid or spot)")
        spot_group.add_argument("--maximum-bid-price-percentage",
                                dest="maximum_bid_price_percentage",
                                type=float,
                                help="maximum value to bid for spot instances" +
                                     " expressed as a percentage of the base" +
                                     " price for the slave node instance type",)
        spot_group.add_argument("--timeout-for-spot-request",
                                dest="timeout_for_request",
                                type=int,
                                help="timeout for a spot instance request" +
                                     " unit: minutes")
        spot_group.add_argument("--maximum-spot-instance-percentage",
                                dest="maximum_spot_instance_percentage",
                                type=int,
                                help="maximum percentage of instances that may" +
                                     " be purchased from the aws spot market," +
                                     " valid only when slave-request-type" +
                                     " is 'hybrid'",)

        stable_spot_group = argparser.add_argument_group("stable spot instance settings")
        stable_spot_group.add_argument("--stable-maximum-bid-price-percentage",
                                       dest="stable_maximum_bid_price_percentage",
                                       type=float,
                                       help="maximum value to bid for stable node spot instances" +
                                       " expressed as a percentage of the base" +
                                       " price for the master and slave node instance types",)
        stable_spot_group.add_argument("--stable-timeout-for-spot-request",
                                       dest="stable_timeout_for_request",
                                       type=int,
                                       help="timeout for a stable node spot instance request" +
                                       " unit: minutes")
        stable_spot_group.add_argument("--stable-allow-fallback",
                                       dest="stable_allow_fallback", default=None,
                                       type=str2bool,
                                       help="whether to fallback to on-demand instances for stable nodes" +
                                       " if spot instances aren't available")

        fairscheduler_group = argparser.add_argument_group(
                              "fairscheduler configuration options")
        fairscheduler_group.add_argument("--fairscheduler-config-xml",
                                         dest="fairscheduler_config_xml_file",
                                         help="location for file containing" +
                                              " xml with custom configuration" +
                                              " for the fairscheduler",)
        fairscheduler_group.add_argument("--fairscheduler-default-pool",
                                         dest="default_pool",
                                         help="default pool for the" +
                                              " fairscheduler",)

        security_group = argparser.add_argument_group("security setttings")
        ephemerals = security_group.add_mutually_exclusive_group()
        ephemerals.add_argument("--encrypted-ephemerals",
                                 dest="encrypted_ephemerals",
                                 action="store_true",
                                 default=None,
                                 help="encrypt the ephemeral drives on" +
                                      " the instance",)
        ephemerals.add_argument("--no-encrypted-ephemerals",
                                 dest="encrypted_ephemerals",
                                 action="store_false",
                                 default=None,
                                 help="don't encrypt the ephemeral drives on" +
                                      " the instance",)
        security_group.add_argument("--customer-ssh-key",
                                    dest="customer_ssh_key_file",
                                    help="location for ssh key to use to" +
                                         " login to the instance")

        presto_group = argparser.add_argument_group("presto settings")
        enabling_presto = presto_group.add_mutually_exclusive_group()
        enabling_presto.add_argument("--enable-presto",
                                  dest="enable_presto",
                                  action="store_true",
                                  default=None,
                                  help="Enable presto for this cluster",)
        enabling_presto.add_argument("--disable-presto",
                                  dest="enable_presto",
                                  action="store_false",
                                  default=None,
                                  help="Disable presto for this cluster",)
        presto_group.add_argument("--presto-custom-config",
                                  dest="presto_custom_config_file",
                                  help="location of file containg custom" +
                                       " presto configuration overrides")

        termination = argparser.add_mutually_exclusive_group()
        termination.add_argument("--disallow-cluster-termination",
                                 dest="disallow_cluster_termination",
                                 action="store_true",
                                 default=None,
                                 help="don't auto-terminate idle clusters," +
                                      " use this with extreme caution",)
        termination.add_argument("--allow-cluster-termination",
                                 dest="disallow_cluster_termination",
                                 action="store_false",
                                 default=None,
                                 help="auto-terminate idle clusters,")

        ganglia = argparser.add_mutually_exclusive_group()
        ganglia.add_argument("--enable-ganglia-monitoring",
                             dest="enable_ganglia_monitoring",
                             action="store_true",
                             default=None,
                             help="enable ganglia monitoring for the" +
                                  " cluster",)
        ganglia.add_argument("--disable-ganglia-monitoring",
                             dest="enable_ganglia_monitoring",
                             action="store_false",
                             default=None,
                             help="disable ganglia monitoring for the" +
                                  " cluster",)

        argparser.add_argument("--node-bootstrap-file",
                dest="node_bootstrap_file",
                help="""name of the node bootstrap file for this cluster. It
                should be in stored in S3 at
                <account-default-location>/scripts/hadoop/NODE_BOOTSTRAP_FILE
                """,)

        argparser.add_argument("--custom-ec2-tags",
                               dest="custom_ec2_tags",
                               help="""Custom ec2 tags to be set on all instances
                               of the cluster. Specified as JSON object (key-value pairs)
                               e.g. --custom-ec2-tags '{"key1":"value1", "key2":"value2"}'
                               """,)

        arguments = argparser.parse_args(args)
        return arguments

    @classmethod
    def create(cls, cluster_info):
        """
        Create a new cluster using information provided in `cluster_info`.
        """
        conn = Qubole.agent()
        return conn.post(cls.rest_entity_path, data=cluster_info)

    @classmethod
    def update(cls, cluster_id_label, cluster_info):
        """
        Update the cluster with id/label `cluster_id_label` using information provided in
        `cluster_info`.
        """
        conn = Qubole.agent()
        return conn.put(cls.element_path(cluster_id_label), data=cluster_info)

    @classmethod
    def clone(cls, cluster_id_label, cluster_info):
        """
        Update the cluster with id/label `cluster_id_label` using information provided in
        `cluster_info`.
        """
        conn = Qubole.agent()
        return conn.post(cls.element_path(cluster_id_label) + '/clone', data=cluster_info)

    @classmethod
    def _parse_reassign_label(cls, args):
        """
        Parse command line arguments for reassigning label.
        """
        argparser = ArgumentParser(prog="cluster reassign_label")

        argparser.add_argument("destination_cluster",
                metavar="destination_cluster_id_label",
                help="id/label of the cluster to move the label to")

        argparser.add_argument("label",
                help="label to be moved from the source cluster")

        arguments = argparser.parse_args(args)
        return arguments

    @classmethod
    def reassign_label(cls, destination_cluster, label):
        """
        Reassign a label from one cluster to another.

        Args:
            `destination_cluster`: id/label of the cluster to move the label to

            `label`: label to be moved from the source cluster
        """
        conn = Qubole.agent()
        data = {
                    "destination_cluster": destination_cluster,
                    "label": label
                }
        return conn.put(cls.rest_entity_path + "/reassign-label", data)

    @classmethod
    def delete(cls, cluster_id_label):
        """
        Delete the cluster with id/label `cluster_id_label`.
        """
        conn = Qubole.agent()
        return conn.delete(cls.element_path(cluster_id_label))


class ClusterInfo():
    """
    qds_sdk.ClusterInfo is the class which stores information about a cluster.
    You can use objects of this class to create or update a cluster.
    """

    def __init__(self, label, aws_access_key_id, aws_secret_access_key,
                 disallow_cluster_termination=None,
                 enable_ganglia_monitoring=None,
                 node_bootstrap_file=None):
        """
        Args:

        `label`: A list of labels that identify the cluster. At least one label
            must be provided when creating a cluster.

        `aws_access_key_id`: The access key id for customer's aws account. This
            is required for creating the cluster.

        `aws_secret_access_key`: The secret access key for customer's aws
            account. This is required for creating the cluster.

        `disallow_cluster_termination`: Set this to True if you don't want
            qubole to auto-terminate idle clusters. Use this option with
            extreme caution.

        `enable_ganglia_monitoring`: Set this to True if you want to enable
            ganglia monitoring for the cluster.

        `node_bootstrap_file`: name of the node bootstrap file for this
            cluster. It should be in stored in S3 at
            <your-default-location>/scripts/hadoop/
        """
        self.label = label
        self.ec2_settings = {}
        self.ec2_settings['compute_access_key'] = aws_access_key_id
        self.ec2_settings['compute_secret_key'] = aws_secret_access_key
        self.disallow_cluster_termination = disallow_cluster_termination
        self.enable_ganglia_monitoring = enable_ganglia_monitoring
        self.node_bootstrap_file = node_bootstrap_file
        self.hadoop_settings = {}
        self.security_settings = {}
        self.presto_settings = {}

    def set_ec2_settings(self,
                         aws_region=None,
                         aws_availability_zone=None,
                         vpc_id=None,
                         subnet_id=None):
        """
        Kwargs:

        `aws_region`: AWS region to create the cluster in.

        `aws_availability_zone`: The availability zone to create the cluster
            in.

        `vpc_id`: The vpc to create the cluster in.

        `subnet_id`: The subnet to create the cluster in.
        """
        self.ec2_settings['aws_region'] = aws_region
        self.ec2_settings['aws_preferred_availability_zone'] = aws_availability_zone
        self.ec2_settings['vpc_id'] = vpc_id
        self.ec2_settings['subnet_id'] = subnet_id

    def set_hadoop_settings(self, master_instance_type=None,
                            slave_instance_type=None,
                            initial_nodes=None,
                            max_nodes=None,
                            custom_config=None,
                            slave_request_type=None,
                            use_hbase=None,
                            custom_ec2_tags=None,
                            use_hadoop2=None):
        """
        Kwargs:

        `master_instance_type`: The instance type to use for the Hadoop master
            node.

        `slave_instance_type`: The instance type to use for the Hadoop slave
            nodes.

        `initial_nodes`: Number of nodes to start the cluster with.

        `max_nodes`: Maximum number of nodes the cluster may be auto-scaled up
            to.

        `custom_config`: Custom Hadoop configuration overrides.

        `slave_request_type`: Purchasing option for slave instances.
            Valid values: "ondemand", "hybrid", "spot".

        `use_hbase`: Start hbase daemons on the cluster. Uses Hadoop2

        `use_hadoop2`: Use hadoop2 in this cluster
        """
        self.hadoop_settings['master_instance_type'] = master_instance_type
        self.hadoop_settings['slave_instance_type'] = slave_instance_type
        self.hadoop_settings['initial_nodes'] = initial_nodes
        self.hadoop_settings['max_nodes'] = max_nodes
        self.hadoop_settings['custom_config'] = custom_config
        self.hadoop_settings['slave_request_type'] = slave_request_type
        self.hadoop_settings['use_hbase'] = use_hbase
        self.hadoop_settings['use_hadoop2'] = use_hadoop2

        if custom_ec2_tags and custom_ec2_tags.strip():
            try:
                self.hadoop_settings['custom_ec2_tags'] = json.loads(custom_ec2_tags.strip())
            except Exception as e:
                raise Exception("Invalid JSON string for custom ec2 tags: %s" % e.message)

    def set_spot_instance_settings(self, maximum_bid_price_percentage=None,
                                   timeout_for_request=None,
                                   maximum_spot_instance_percentage=None):
        """
        Purchase options for spot instances. Valid only when
        `slave_request_type` is hybrid or spot.

        `maximum_bid_price_percentage`: Maximum value to bid for spot
            instances, expressed as a percentage of the base price for the
            slave node instance type.

        `timeout_for_request`: Timeout for a spot instance request (Unit:
            minutes)

        `maximum_spot_instance_percentage`: Maximum percentage of instances
            that may be purchased from the AWS Spot market. Valid only when
            slave_request_type is "hybrid".
        """
        self.hadoop_settings['spot_instance_settings'] = {
               'maximum_bid_price_percentage': maximum_bid_price_percentage,
               'timeout_for_request': timeout_for_request,
               'maximum_spot_instance_percentage': maximum_spot_instance_percentage}


    def set_stable_spot_instance_settings(self, maximum_bid_price_percentage=None,
                                          timeout_for_request=None,
                                          allow_fallback=True):
        """
        Purchase options for stable spot instances.

        `maximum_bid_price_percentage`: Maximum value to bid for stable node spot
            instances, expressed as a percentage of the base price
            (applies to both master and slave nodes).

        `timeout_for_request`: Timeout for a stable node spot instance request (Unit:
            minutes)

        `allow_fallback`: Whether to fallback to on-demand instances for
            stable nodes if spot instances are not available
        """
        self.hadoop_settings['stable_spot_instance_settings'] = {
               'maximum_bid_price_percentage': maximum_bid_price_percentage,
               'timeout_for_request': timeout_for_request,
               'allow_fallback': allow_fallback}


    def set_fairscheduler_settings(self, fairscheduler_config_xml=None,
                                   default_pool=None):
        """
        Fair scheduler configuration options.

        `fairscheduler_config_xml`: XML string with custom configuration
            parameters for the fair scheduler.

        `default_pool`: The default pool for the fair scheduler.
        """
        self.hadoop_settings['fairscheduler_settings'] = {
               'fairscheduler_config_xml': fairscheduler_config_xml,
               'default_pool': default_pool}

    def set_security_settings(self,
                              encrypted_ephemerals=None,
                              customer_ssh_key=None):
        """
        Kwargs:

        `encrypted_ephemerals`: Encrypt the ephemeral drives on the instance.

        `customer_ssh_key`: SSH key to use to login to the instances.
        """
        self.security_settings['encrypted_ephemerals'] = encrypted_ephemerals
        self.security_settings['customer_ssh_key'] = customer_ssh_key

    def set_presto_settings(self, enable_presto=None, presto_custom_config=None):
        """
        Kwargs:

        `enable_presto`: Enable Presto on the cluster.

        `presto_custom_config`: Custom Presto configuration overrides.
        """
        self.presto_settings['enable_presto'] = enable_presto
        self.presto_settings['custom_config'] = presto_custom_config

    def minimal_payload(self):
        """
        This method can be used to create the payload which is sent while
        creating or updating a cluster.
        """
        payload = {"cluster": self.__dict__}
        return _make_minimal(payload)


def _make_minimal(dictionary):
    """
    This function removes all the keys whose value is either None or an empty
    dictionary.
    """
    new_dict = {}
    for key, value in dictionary.items():
        if value is not None:
            if isinstance(value, dict):
                new_value = _make_minimal(value)
                if new_value:
                    new_dict[key] = new_value
            else:
                new_dict[key] = value
    return new_dict



import os
import sys
from setuptools import setup

INSTALL_REQUIRES = ['requests >=1.0.3', 'boto >=2.1.1', 'six >=1.2.0', 'urllib3 >= 1.0.2']
if sys.version_info < (2, 7, 0):
    INSTALL_REQUIRES.append('argparse>=1.1')


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name="qds_sdk",
    version="1.5.0",
    author="Qubole",
    author_email="dev@qubole.com",
    description=("Python SDK for coding to the Qubole Data Service API"),
    keywords="qubole sdk api",
    url="https://github.com/qubole/qds-sdk-py",
    packages=['qds_sdk'],
    scripts=['bin/qds.py'],
    install_requires=INSTALL_REQUIRES,
    long_description=read('README.rst'),
    classifiers=[
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4"
    ]
    )



import json

from qds_sdk.qubole import Qubole
from qds_sdk.resource import Resource
from argparse import ArgumentParser
from qds_sdk.commands import *
from qds_sdk.actions import *

class SchedulerCmdLine:
    """
    qds_sdk.ScheduleCmdLine is the interface used by qds.py.
    """

    @staticmethod
    def parsers():
        argparser = ArgumentParser(prog="qds.py scheduler",
                                        description="Scheduler client for Qubole Data Service.")
        subparsers = argparser.add_subparsers()

        #Create
        create = subparsers.add_parser("create",
                                       help="Create a new schedule")
        create.add_argument("--data", dest="data", required=True,
                            help="Path to a file that contains scheduler attributes as a json object")
        create.set_defaults(func=SchedulerCmdLine.create)

        #List
        list = subparsers.add_parser("list",
                                     help="List all schedulers")
        list.add_argument("--fields", nargs="*", dest="fields",
                          help="List of fields to show")
        list.add_argument("--per-page", dest="per_page",
                          help="Number of items per page")
        list.add_argument("--page", dest="page",
                          help="Page Number")
        list.set_defaults(func=SchedulerCmdLine.list)

        #View
        view = subparsers.add_parser("view",
                                     help="View a specific schedule")
        view.add_argument("id", help="Numeric id of the schedule")
        view.add_argument("--fields", nargs="*", dest="fields",
                          help="List of fields to show")
        view.set_defaults(func=SchedulerCmdLine.view)

        #View by name
        view_by_name = subparsers.add_parser("view_by_name",
                                     help="View a specific schedule")
        view_by_name.add_argument("name", help="Name of the schedule")
        view_by_name.add_argument("--fields", nargs="*", dest="fields",
                          help="List of fields to show")
        view_by_name.set_defaults(func=SchedulerCmdLine.view_by_name)

        #Suspend
        suspend = subparsers.add_parser("suspend",
                                        help="Suspend a specific schedule")
        suspend.add_argument("id", help="Numeric id or name of the schedule")
        suspend.set_defaults(func=SchedulerCmdLine.suspend)

        #Resume
        resume = subparsers.add_parser("resume",
                                       help="Resume a specific schedule")
        resume.add_argument("id", help="Numeric id or name of the schedule")
        resume.set_defaults(func=SchedulerCmdLine.resume)

        #Kill
        kill = subparsers.add_parser("kill",
                                     help="Kill a specific schedule")
        kill.add_argument("id", help="Numeric id or name of the schedule")
        kill.set_defaults(func=SchedulerCmdLine.kill)

        #List Actions
        list_actions = subparsers.add_parser("list-actions",
                                               help="List actions of a specific schedule")
        list_actions.add_argument("id", help="Numeric id or name of the schedule")
        list_actions.add_argument("--sequence_id", dest="sequence_id", help="Sequence id of the actions to list")
        list_actions.add_argument("--fields", nargs="*", dest="fields",
                          help="List of fields to show")
        list_actions.add_argument("--per-page", dest="per_page",
                                    help="Number of items per page")
        list_actions.add_argument("--page", dest="page",
                                    help="Page Number")
        list_actions.set_defaults(func=SchedulerCmdLine.list_actions)

        #List Instances
        list_instances = subparsers.add_parser("list-instances",
                                               help="List instances of a specific schedule")
        list_instances.add_argument("id", help="Numeric id or name of the schedule")
        list_instances.add_argument("--fields", nargs="*", dest="fields",
                          help="List of fields to show")
        list_instances.add_argument("--per-page", dest="per_page",
                                    help="Number of items per page")
        list_instances.add_argument("--page", dest="page",
                                    help="Page Number")
        list_instances.set_defaults(func=SchedulerCmdLine.list_instances)

        rerun = subparsers.add_parser("rerun",
                                      help="Rerun an instance of a schedule")
        rerun.add_argument("id", help="Numeric id or name of the schedule")
        rerun.add_argument("instance_id", help="Numeric id of the instance")
        rerun.set_defaults(func=SchedulerCmdLine.rerun)
        return argparser

    @staticmethod
    def run(args):
        parser = SchedulerCmdLine.parsers()
        parsed = parser.parse_args(args)
        return parsed.func(parsed)

    @staticmethod
    def filter_fields(schedule, fields):
        filtered = {}
        for field in fields:
            filtered[field] = schedule[field]
        return filtered

    @staticmethod
    def create(args):
        with open(args.data) as f:
            spec = json.load(f)
        schedule = Scheduler(spec)
        return json.dumps(schedule.attributes, sort_keys=True, indent=4)

    @staticmethod
    def list(args):
        schedlist = Scheduler.list(args.page, args.per_page)
        if args.fields:
            for s in schedlist:
                s.attributes = SchedulerCmdLine.filter_fields(s.attributes, args.fields)
        return json.dumps(schedlist, default=lambda o: o.attributes, sort_keys=True, indent=4)

    @staticmethod
    def view(args):
        schedule = Scheduler.find(args.id)
        if args.fields:
            schedule.attributes = SchedulerCmdLine.filter_fields(schedule.attributes, args.fields)
        return json.dumps(schedule.attributes, sort_keys=True, indent=4)

    @staticmethod
    def view_by_name(args):
        schedule = Scheduler.find_by_name(args.name)
        if schedule is None:
            return "Schedule '%s' not found" % args.name
        if args.fields:
            schedule.attributes = SchedulerCmdLine.filter_fields(schedule.attributes, args.fields)
        return json.dumps(schedule.attributes, sort_keys=True, indent=4)

    @staticmethod
    def suspend(args):
        schedule = Scheduler.find(args.id)
        return json.dumps(schedule.suspend(), sort_keys=True, indent=4)

    @staticmethod
    def resume(args):
        schedule = Scheduler.find(args.id)
        return json.dumps(schedule.resume(), sort_keys=True, indent=4)

    @staticmethod
    def kill(args):
        schedule = Scheduler.find(args.id)
        return json.dumps(schedule.kill(), sort_keys=True, indent=4)

    @staticmethod
    def list_actions(args):
        schedule = Scheduler.find(args.id)
        actlist = schedule.list_actions(args.sequence_id, args.page, args.per_page)
        if args.fields:
            for a in actlist:
                a.attributes = ActionCmdLine.filter_fields(a.attributes, args.fields)
        return json.dumps(actlist, default=lambda o: o.attributes,
                          sort_keys=True, indent=4)

    @staticmethod
    def list_instances(args):
        schedule = Scheduler.find(args.id)
        cmdlist = schedule.list_instances(args.page, args.per_page)
        if args.fields:
            for cmd in cmdlist:
                cmd.attributes = SchedulerCmdLine.filter_fields(cmd.attributes, args.fields)
        return json.dumps(cmdlist, default=lambda o: o.attributes,
                          sort_keys=True, indent=4)

    @staticmethod
    def rerun(args):
        schedule = Scheduler.find(args.id)
        return schedule.rerun(args.instance_id)


class Scheduler(Resource):
    """
    qds_sdk.Schedule is the base Qubole Schedule class.
    """

    """ all commands use the /scheduler endpoint"""

    rest_entity_path = "scheduler"

    @staticmethod
    def list(page = None, per_page = None):
        conn = Qubole.agent()
        url_path = Scheduler.rest_entity_path
        page_attr = []
        if page is not None:
            page_attr.append("page=%s" % page)
        if per_page is not None:
            page_attr.append("per_page=%s" % per_page)
        if page_attr:
            url_path = "%s?%s" % (Scheduler.rest_entity_path, "&".join(page_attr))

        #Todo Page numbers are thrown away right now
        schedjson = conn.get(url_path)
        schedlist = []
        for s in schedjson["schedules"]:
            schedlist.append(Scheduler(s))
        return schedlist

    @staticmethod
    def find_by_name(name):
        conn = Qubole.agent()
        if name is not None:
            schedjson = conn.get(Scheduler.rest_entity_path, params={"name":name})
            if schedjson["schedules"]:
                return Scheduler(schedjson["schedules"][0])
        return None

    def suspend(self):
        conn = Qubole.agent()
        data = {"status": "suspend"}
        return conn.put(self.element_path(self.id), data)

    def resume(self):
        conn = Qubole.agent()
        data = {"status": "resume"}
        return conn.put(self.element_path(self.id), data)

    def kill(self):
        conn = Qubole.agent()
        data = {"status": "kill"}
        return conn.put(self.element_path(self.id), data)

    def list_actions(self, sequence_id = None, page=None, per_page=None):
        conn = Qubole.agent()
        url_path = self.element_path(self.id) + "/" + "actions"
        if sequence_id is not None:
            url_path = url_path + "/" +  str(sequence_id)
        params = {}
        if page is not None:
            params['page'] = page
        if per_page is not None:
            params['per_page'] = per_page

        #Todo Page numbers are thrown away right now
        actjson = conn.get(url_path, params)
        actlist = []
        for act in actjson["actions"]:
            actlist.append(Action(act))
        return actlist

    def list_instances(self, page=None, per_page=None):
        conn = Qubole.agent()
        url_path = self.element_path(self.id) + "/" + "instances"
        page_attr = []
        if page is not None:
            page_attr.append("page=%s" % page)
        if per_page is not None:
            page_attr.append("per_page=%s" % per_page)
        if page_attr:
            url_path = "%s/instances?%s" % (self.element_path(args.id), "&".join(page_attr))
        #Todo Page numbers are thrown away right now
        cmdjson = conn.get(url_path)
        cmdlist = []
        for cmd in cmdjson["commands"]:
            cmdclass = globals()[cmd["command_type"]]
            onecmd = cmdclass(cmd)
            cmdlist.append(onecmd)
        return cmdlist

    def rerun(self, instance_id):
        conn = Qubole.agent()
        url_path = self.element_path(id) + "/instances/" + instance_id + "/rerun"
        return conn.post(url_path)['status']



"""
The commands module contains the base definition for
a generic Qubole command and the implementation of all
the specific commands
"""

from qds_sdk.qubole import Qubole
from qds_sdk.resource import Resource
from qds_sdk.exception import ParseError
from qds_sdk.account import Account
from qds_sdk.util import GentleOptionParser
from qds_sdk.util import OptionParsingError
from qds_sdk.util import OptionParsingExit
from optparse import SUPPRESS_HELP
import boto

import time
import logging
import sys
import re
import pipes
import os
import json

log = logging.getLogger("qds_commands")

# Pattern matcher for s3 path
_URI_RE = re.compile(r's3://([^/]+)/?(.*)')


class Command(Resource):

    """
    qds_sdk.Command is the base Qubole command class. Different types of Qubole
    commands can subclass this.
    """

    """all commands use the /commands endpoint"""
    rest_entity_path = "commands"

    @staticmethod
    def is_done(status):
        """
        Does the status represent a completed command
        Args:
            `status`: a status string

        Returns:
            True/False
        """
        return status == "cancelled" or status == "done" or status == "error"

    @staticmethod
    def is_success(status):
        return status == "done"

    @classmethod
    def create(cls, **kwargs):
        """
        Create a command object by issuing a POST request to the /command endpoint
        Note - this does not wait for the command to complete

        Args:
            `**kwargs`: keyword arguments specific to command type

        Returns:
            Command object
        """

        conn = Qubole.agent()
        if kwargs.get('command_type') is None:
            kwargs['command_type'] = cls.__name__
        if kwargs.get('tags') is not None:
            kwargs['tags'] = kwargs['tags'].split(',')

        return cls(conn.post(cls.rest_entity_path, data=kwargs))

    @classmethod
    def run(cls, **kwargs):
        """
        Create a command object by issuing a POST request to the /command endpoint
        Waits until the command is complete. Repeatedly polls to check status

        Args:
            `**kwargs`: keyword arguments specific to command type

        Returns:
            Command object
        """
        cmd = cls.create(**kwargs)
        while not Command.is_done(cmd.status):
            time.sleep(Qubole.poll_interval)
            cmd = cls.find(cmd.id)

        return cmd

    @classmethod
    def cancel_id(cls, id):
        """
        Cancels command denoted by this id

        Args:
            `id`: command id
        """
        conn = Qubole.agent()
        data = {"status": "kill"}
        return conn.put(cls.element_path(id), data)

    def cancel(self):
        """
        Cancels command represented by this object
        """
        self.__class__.cancel_id(self.id)

    @classmethod
    def get_log_id(cls, id):
        """
        Fetches log for the command represented by this id

        Args:
            `id`: command id
        """
        conn = Qubole.agent()
        r = conn.get_raw(cls.element_path(id) + "/logs")
        return r.text

    def get_log(self):
        """
        Fetches log for the command represented by this object

        Returns:
            The log as a string
        """
        log_path = self.meta_data['logs_resource']
        conn = Qubole.agent()
        r = conn.get_raw(log_path)
        return r.text


    @classmethod
    def get_jobs_id(cls, id):
        """
        Fetches information about the hadoop jobs which were started by this
        command id. This information is only available for commands which have
        completed (i.e. Status = 'done', 'cancelled' or 'error'.) Also, the
        cluster which ran this command should be running for this information
        to be available. Otherwise only the URL and job_id is shown.

        Args:
            `id`: command id
        """
        conn = Qubole.agent()
        r = conn.get_raw(cls.element_path(id) + "/jobs")
        return r.text


    def get_results(self, fp=sys.stdout, inline=True, delim=None, fetch=True):
        """
        Fetches the result for the command represented by this object

        get_results will retrieve results of the command and write to stdout by default.
        Optionally one can write to a filestream specified in `fp`. The `inline` argument
        decides whether the result can be returned as a CRLF separated string. In cases where
        the results are greater than 20MB, get_results will attempt to read from s3 and write
        to fp. The retrieval of results from s3 can be turned off by the `fetch` argument

        Args:
            `fp`: a file object to write the results to directly
            `inline`: whether or not results are returned inline as CRLF separated string
            `fetch`: True to fetch the result even if it is greater than 20MB, False to
                     only get the result location on s3
        """
        result_path = self.meta_data['results_resource']

        conn = Qubole.agent()

        r = conn.get(result_path, {'inline': inline})
        if r.get('inline'):
            if sys.version_info < (3, 0, 0):
                fp.write(r['results'].encode('utf8'))
            else:
                import io
                if isinstance(fp, io.TextIOBase):
                    fp.buffer.write(r['results'].encode('utf8'))
                elif isinstance(fp, io.BufferedIOBase) or isinstance(fp, io.RawIOBase):
                    fp.write(r['results'].encode('utf8'))
                else:
                    # Can this happen? Don't know what's the right thing to do in this case.
                    pass
        else:
            if fetch:
                acc = Account.find()
                boto_conn = boto.connect_s3(aws_access_key_id=acc.storage_access_key,
                                            aws_secret_access_key=acc.storage_secret_key)

                log.info("Starting download from result locations: [%s]" % ",".join(r['result_location']))
                #fetch latest value of num_result_dir
                num_result_dir = Command.find(self.id).num_result_dir
                for s3_path in r['result_location']:
                    # In Python 3,
                    # If the delim is None, fp should be in binary mode because
                    # boto expects it to be.
                    # If the delim is not None, then both text and binary modes
                    # work.
                    _download_to_local(boto_conn, s3_path, fp, num_result_dir, delim=delim)
            else:
                fp.write(",".join(r['result_location']))



class HiveCommand(Command):

    usage = ("hivecmd <submit|run> [options]")

    optparser = GentleOptionParser(usage=usage)
    optparser.add_option("-q", "--query", dest="query", help="query string")

    optparser.add_option("-f", "--script_location", dest="script_location",
                         help="Path where hive query to run is stored. Can be S3 URI or local file path")

    optparser.add_option("--macros", dest="macros",
                         help="expressions to expand macros used in query")

    optparser.add_option("--tags", dest="tags",
                         help="comma-separated list of tags to be associated with the query ( e.g., tag1 tag1,tag2 )")

    optparser.add_option("--sample_size", dest="sample_size",
                         help="size of sample in bytes on which to run query")

    optparser.add_option("--cluster-label", dest="label",
                         help="the label of the cluster to run the command on")

    optparser.add_option("--notify", action="store_true", dest="can_notify",
                         default=False, help="sends an email on command completion")

    optparser.add_option("--name", dest="name",
                         help="Assign a name to this query")

    optparser.add_option("--print-logs", action="store_true", dest="print_logs",
                         default=False, help="Fetch logs and print them to stderr.")

    @classmethod
    def parse(cls, args):
        """
        Parse command line arguments to construct a dictionary of command
        parameters that can be used to create a command

        Args:
            `args`: sequence of arguments

        Returns:
            Dictionary that can be used in create method

        Raises:
            ParseError: when the arguments are not correct
        """

        try:
            (options, args) = cls.optparser.parse_args(args)
            if options.query is None and options.script_location is None:
                raise ParseError("One of query or script location"
                                 " must be specified",
                                 cls.optparser.format_help())
        except OptionParsingError as e:
            raise ParseError(e.msg, cls.optparser.format_help())
        except OptionParsingExit as e:
            return None

        if options.script_location is not None:
            if options.query is not None:
                raise ParseError(
                    "Both query and script_location cannot be specified",
                    cls.optparser.format_help())

            if ((options.script_location.find("s3://") != 0) and
                (options.script_location.find("s3n://") != 0)):

                # script location is local file

                try:
                    q = open(options.script_location).read()
                except IOError as e:
                    raise ParseError("Unable to open script location: %s" %
                                     str(e),
                                     cls.optparser.format_help())
                options.script_location = None
                options.query = q

        if options.macros is not None:
            options.macros = json.loads(options.macros)
        v = vars(options)
        v["command_type"] = "HiveCommand"
        return v

class SparkCommand(Command):

    usage = ("sparkcmd <submit|run> [options]")
    allowedlanglist = ["python", "scala"]

    optparser = GentleOptionParser(usage=usage)
    optparser.add_option("--program", dest="program",help=SUPPRESS_HELP)

    optparser.add_option("--cmdline", dest="cmdline", help="command line for Spark")

    optparser.add_option("-f", "--script_location", dest="script_location",
                         help="Path where spark program to run is stored. Has to be a local file path")

    optparser.add_option("--macros", dest="macros",
                         help="expressions to expand macros used in query")

    optparser.add_option("--tags", dest="tags",
                         help="comma-separated list of tags to be associated with the query ( e.g., tag1 tag1,tag2 )")

    optparser.add_option("--cluster-label", dest="label", help="the label of the cluster to run the command on")

    optparser.add_option("--language", dest="language", choices = allowedlanglist, help=SUPPRESS_HELP)

    optparser.add_option("--notify", action="store_true", dest="can_notify", default=False, help="sends an email on command completion")

    optparser.add_option("--name", dest="name", help="Assign a name to this query")

    optparser.add_option("--arguments", dest = "arguments", help = "Spark Submit Command Line Options")

    optparser.add_option("--user_program_arguments", dest = "user_program_arguments", help = "Arguments for User Program")

    optparser.add_option("--print-logs", action="store_true", dest="print_logs",
                         default=False, help="Fetch logs and print them to stderr.")
    @classmethod
    def validate_program(cls, options):
        bool_program = options.program is not None
        bool_other_options = options.script_location is not None or options.cmdline is not None

        # if both are false then no option is specified ==> raise ParseError
        # if both are true then atleast two option specified ==> raise ParseError
        if bool_program == bool_other_options:
            raise ParseError("Exactly One of script location or program or cmdline should be specified", cls.optparser.format_help())
        if bool_program:
            if options.language is None:
                raise ParseError("Unspecified language for Program", cls.optparser.format_help())

    @classmethod
    def validate_cmdline(cls, options):
        bool_cmdline = options.cmdline is not None
        bool_other_options = options.script_location is not None or options.program is not None

        # if both are false then no option is specified ==> raise ParseError
        # if both are true then atleast two option specified ==> raise ParseError
        if bool_cmdline == bool_other_options:
            raise ParseError("Exactly One of script location or program or cmdline should be specified", cls.optparser.format_help())
        if bool_cmdline:
            if options.language is not None:
                raise ParseError("Language cannot be specified with the commandline option", cls.optparser.format_help())

    @classmethod
    def validate_script_location(cls, options):
        bool_script_location = options.script_location is not None
        bool_other_options = options.program is not None or options.cmdline is not None

        # if both are false then no option is specified ==> raise ParseError
        # if both are true then atleast two option specified ==> raise ParseError
        if bool_script_location == bool_other_options:
            raise ParseError("Exactly One of script location or program or cmdline should be specified", cls.optparser.format_help())

        if bool_script_location:
            if options.language is not None:
                raise ParseError("Both script location and language cannot be specified together", cls.optparser.format_help())
            # for now, aws script_location is not supported and throws an error
            if ((options.script_location.find("s3://") != 0) and
                (options.script_location.find("s3n://") != 0)):

                # script location is local file so set the program as the text from the file

                try:
                    q = open(options.script_location).read()
                except IOError as e:
                    raise ParseError("Unable to open script location: %s" %
                                     str(e),
                                     cls.optparser.format_help())


                fileName, fileExtension = os.path.splitext(options.script_location)
                # getting the language of the program from the file extension
                if fileExtension == ".py":
                    options.language = "python"
                elif fileExtension == ".scala":
                    options.language = "scala"
                else:
                    raise ParseError("Invalid program type, Please choose one from python or scala %s" %str(fileExtension),
                                     cls.optparser.format_help())
            else:
                raise ParseError("Invalid location, Please choose a local file location",
                                 cls.optparser.format_help())

            options.script_location = None
            options.program = q

    @classmethod
    def parse(cls, args):
        """
        Parse command line arguments to construct a dictionary of command
        parameters that can be used to create a command

        Args:
            `args`: sequence of arguments

        Returns:
            Dictionary that can be used in create method

        Raises:
            ParseError: when the arguments are not correct
        """
        try:
            (options, args) = cls.optparser.parse_args(args)
        except OptionParsingError as e:
            raise ParseError(e.msg, cls.optparser.format_help())
        except OptionParsingExit as e:
            return None

        SparkCommand.validate_program(options)
        SparkCommand.validate_script_location(options)
        SparkCommand.validate_cmdline(options)

        if options.macros is not None:
            options.macros = json.loads(options.macros)

        v = vars(options)
        v["command_type"] = "SparkCommand"
        return v



class PrestoCommand(Command):

    usage = ("prestocmd <submit|run> [options]")

    optparser = GentleOptionParser(usage=usage)
    optparser.add_option("-q", "--query", dest="query", help="query string")

    optparser.add_option("-f", "--script_location", dest="script_location",
                         help="Path where presto query to run is stored. Can be S3 URI or local file path")

    optparser.add_option("--macros", dest="macros",
                         help="expressions to expand macros used in query")

    optparser.add_option("--tags", dest="tags",
                         help="comma-separated list of tags to be associated with the query ( e.g., tag1 tag1,tag2 )")

    optparser.add_option("--cluster-label", dest="label",
                         help="the label of the cluster to run the command on")

    optparser.add_option("--notify", action="store_true", dest="can_notify",
                         default=False, help="sends an email on command completion")

    optparser.add_option("--name", dest="name",
                         help="Assign a name to this query")

    optparser.add_option("--print-logs", action="store_true", dest="print_logs",
                         default=False, help="Fetch logs and print them to stderr.")

    @classmethod
    def parse(cls, args):
        """
        Parse command line arguments to construct a dictionary of command
        parameters that can be used to create a command

        Args:
            `args`: sequence of arguments

        Returns:
            Dictionary that can be used in create method

        Raises:
            ParseError: when the arguments are not correct
        """

        try:
            (options, args) = cls.optparser.parse_args(args)
            if options.query is None and options.script_location is None:
                raise ParseError("One of query or script location"
                                 " must be specified",
                                 cls.optparser.format_help())
        except OptionParsingError as e:
            raise ParseError(e.msg, cls.optparser.format_help())
        except OptionParsingExit as e:
            return None

        if options.script_location is not None:
            if options.query is not None:
                raise ParseError(
                    "Both query and script_location cannot be specified",
                    cls.optparser.format_help())

            if ((options.script_location.find("s3://") != 0) and
                (options.script_location.find("s3n://") != 0)):

                # script location is local file
                try:
                    q = open(options.script_location).read()
                except IOError as e:
                    raise ParseError("Unable to open script location: %s" %
                                     str(e),
                                     cls.optparser.format_help())
                options.script_location = None
                options.query = q

        if options.macros is not None:
            options.macros = json.loads(options.macros)
        v = vars(options)
        v["command_type"] = "PrestoCommand"
        return v


class HadoopCommand(Command):
    subcmdlist = ["jar", "s3distcp", "streaming"]
    usage = "hadoopcmd <submit|run> [options] <%s> <arg1> [arg2] ..." % "|".join(subcmdlist)

    optparser = GentleOptionParser(usage=usage)
    optparser.add_option("--cluster-label", dest="label",
                         help="the label of the cluster to run the command on")

    optparser.add_option("--notify", action="store_true", dest="can_notify",
                         default=False, help="sends an email on command completion")

    optparser.add_option("--name", dest="name",
                         help="Assign a name to this command")

    optparser.add_option("--tags", dest="tags",
                         help="comma-separated list of tags to be associated with the query ( e.g., tag1 tag1,tag2 )")

    optparser.add_option("--print-logs", action="store_true", dest="print_logs",
                         default=False, help="Fetch logs and print them to stderr.")

    optparser.disable_interspersed_args()

    @classmethod
    def parse(cls, args):
        """
        Parse command line arguments to construct a dictionary of command
        parameters that can be used to create a command

        Args:
            `args`: sequence of arguments

        Returns:
            Dictionary that can be used in create method

        Raises:
            ParseError: when the arguments are not correct
        """
        parsed = {}

        try:
            (options, args) = cls.optparser.parse_args(args)
        except OptionParsingError as e:
            raise ParseError(e.msg, cls.optparser.format_help())
        except OptionParsingExit as e:
            return None

        parsed['label'] = options.label
        parsed['can_notify'] = options.can_notify
        parsed['name'] = options.name
        parsed['tags'] = options.tags
        parsed["command_type"] = "HadoopCommand"
        parsed['print_logs'] = options.print_logs

        if len(args) < 2:
            raise ParseError("Need at least two arguments", cls.usage)

        subcmd = args.pop(0)
        if subcmd not in cls.subcmdlist:
            raise ParseError("First argument must be one of <%s>" %
                             "|".join(cls.subcmdlist))

        parsed["sub_command"] = subcmd
        parsed["sub_command_args"] = " ".join("'" + a + "'" for a in args)

        return parsed


class ShellCommand(Command):
    usage = ("shellcmd <submit|run> [options] [arg1] [arg2] ...")

    optparser = GentleOptionParser(usage=usage)
    optparser.add_option("-s", "--script", dest="inline", help="inline script that can be executed by bash")

    optparser.add_option("-f", "--script_location", dest="script_location",
                         help="Path where bash script to run is stored. Can be S3 URI or local file path")

    optparser.add_option("-i", "--files", dest="files",
                         help="List of files [optional] Format : file1,file2 (files in s3 bucket) These files will be copied to the working directory where the command is executed")

    optparser.add_option("-a", "--archives", dest="archives",
                         help="List of archives [optional] Format : archive1,archive2 (archives in s3 bucket) These are unarchived in the working directory where the command is executed")

    optparser.add_option("--cluster-label", dest="label",
                         help="the label of the cluster to run the command on")

    optparser.add_option("--notify", action="store_true", dest="can_notify",
                         default=False, help="sends an email on command completion")

    optparser.add_option("--tags", dest="tags",
                         help="comma-separated list of tags to be associated with the query ( e.g., tag1 tag1,tag2 )")

    optparser.add_option("--name", dest="name",
                         help="Assign a name to this command")

    optparser.add_option("--print-logs", action="store_true", dest="print_logs",
                         default=False, help="Fetch logs and print them to stderr.")

    @classmethod
    def parse(cls, args):
        """
        Parse command line arguments to construct a dictionary of command
        parameters that can be used to create a command

        Args:
            `args`: sequence of arguments

        Returns:
            Dictionary that can be used in create method

        Raises:
            ParseError: when the arguments are not correct
        """

        try:
            (options, args) = cls.optparser.parse_args(args)
            if options.inline is None and options.script_location is None:
                raise ParseError("One of script or it's location"
                                 " must be specified",
                                 cls.optparser.format_help())
        except OptionParsingError as e:
            raise ParseError(e.msg, cls.optparser.format_help())
        except OptionParsingExit as e:
            return None

        if options.script_location is not None:
            if options.inline is not None:
                raise ParseError(
                    "Both script and script_location cannot be specified",
                    cls.optparser.format_help())

            if ((options.script_location.find("s3://") != 0) and
                (options.script_location.find("s3n://") != 0)):

                # script location is local file

                try:
                    s = open(options.script_location).read()
                except IOError as e:
                    raise ParseError("Unable to open script location: %s" %
                                     str(e),
                                     cls.optparser.format_help())
                options.script_location = None
                options.inline = s

            if (args is not None) and (len(args) > 0):
                if options.inline is not None:
                    raise ParseError(
                        "Extra arguments can only be "
                        "supplied with a script_location in S3 right now",
                        cls.optparser.format_help())

                setattr(options, 'parameters',
                        " ".join([pipes.quote(a) for a in args]))

        else:
            if (args is not None) and (len(args) > 0):
                raise ParseError(
                    "Extra arguments can only be supplied with a script_location",
                    cls.optparser.format_help())

        v = vars(options)
        v["command_type"] = "ShellCommand"
        return v


class PigCommand(Command):
    usage = ("pigcmd <submit|run> [options] [key1=value1] [key2=value2] ...")

    optparser = GentleOptionParser(usage=usage)
    optparser.add_option("-s", "--script", dest="latin_statements",
                         help="latin statements that has to be executed")

    optparser.add_option("-f", "--script_location", dest="script_location",
                         help="Path where bash script to run is stored. Can be S3 URI or local file path")

    optparser.add_option("--cluster-label", dest="label",
                         help="the label of the cluster to run the command on")

    optparser.add_option("--notify", action="store_true", dest="can_notify",
                         default=False, help="sends an email on command completion")

    optparser.add_option("--tags", dest="tags",
                         help="comma-separated list of tags to be associated with the query ( e.g., tag1 tag1,tag2 )")

    optparser.add_option("--name", dest="name",
                         help="Assign a name to this command")

    optparser.add_option("--print-logs", action="store_true", dest="print_logs",
                         default=False, help="Fetch logs and print them to stderr.")

    @classmethod
    def parse(cls, args):
        """
        Parse command line arguments to construct a dictionary of command
        parameters that can be used to create a command

        Args:
            `args`: sequence of arguments

        Returns:
            Dictionary that can be used in create method

        Raises:
            ParseError: when the arguments are not correct
        """

        try:
            (options, args) = cls.optparser.parse_args(args)
            if options.latin_statements is None and options.script_location is None:
                raise ParseError("One of script or it's location"
                                 " must be specified",
                                 cls.optparser.format_help())
        except OptionParsingError as e:
            raise ParseError(e.msg, cls.optparser.format_help())
        except OptionParsingExit as e:
            return None

        if options.script_location is not None:
            if options.latin_statements is not None:
                raise ParseError(
                    "Both script and script_location cannot be specified",
                    cls.optparser.format_help())

            if ((options.script_location.find("s3://") != 0) and
                (options.script_location.find("s3n://") != 0)):

                # script location is local file

                try:
                    s = open(options.script_location).read()
                except IOError as e:
                    raise ParseError("Unable to open script location: %s" %
                                     str(e),
                                     cls.optparser.format_help())
                options.script_location = None
                options.latin_statements = s

            if (args is not None) and (len(args) > 0):
                if options.latin_statements is not None:
                    raise ParseError(
                        "Extra arguments can only be "
                        "supplied with a script_location in S3 right now",
                        cls.optparser.format_help())

                p = {}
                for a in args:
                    kv = a.split('=')
                    if len(kv) != 2:
                        raise ParseError("Arguments to pig script must be of this format k1=v1 k2=v2 k3=v3...")
                    p[kv[0]] = kv[1]
                setattr(options, 'parameters', p)

        else:
            if (args is not None) and (len(args) > 0):
                raise ParseError(
                    "Extra arguments can only be supplied with a script_location",
                    cls.optparser.format_help())

        v = vars(options)
        v["command_type"] = "PigCommand"
        return v


class DbExportCommand(Command):
    usage = ("dbexportcmd <submit|run> [options]")

    optparser = GentleOptionParser(usage=usage)
    optparser.add_option("-m", "--mode", dest="mode",
                         help="Can be 1 for Hive export or 2 for HDFS/S3 export")
    optparser.add_option("--hive_table", dest="hive_table",
                         help="Mode 1: Name of the Hive Table from which data will be exported")
    optparser.add_option("--partition_spec", dest="partition_spec",
                         help="Mode 1: (optional) Partition specification for Hive table")
    optparser.add_option("--dbtap_id", dest="dbtap_id",
                         help="Modes 1 and 2: DbTap Id of the target database in Qubole")
    optparser.add_option("--db_table", dest="db_table",
                         help="Modes 1 and 2: Table to export to in the target database")
    optparser.add_option("--db_update_mode", dest="db_update_mode",
                         help="Modes 1 and 2: (optional) can be 'allowinsert' or "
                              "'updateonly'. If updateonly is "
                              "specified - only existing rows are updated. If allowinsert "
                              "is specified - then existing rows are updated and non existing "
                              "rows are inserted. If this option is not specified - then the "
                              "given the data will be appended to the table")
    optparser.add_option("--db_update_keys", dest="db_update_keys",
                         help="Modes 1 and 2: Columns used to determine the uniqueness of rows for "
                              "'updateonly' mode")
    optparser.add_option("--export_dir", dest="export_dir",
                         help="Mode 2: HDFS/S3 location from which data will be exported")
    optparser.add_option("--fields_terminated_by", dest="fields_terminated_by",
                         help="Mode 2: Hex of the char used as column separator "
                              "in the dataset, for eg. \0x20 for space")

    optparser.add_option("--notify", action="store_true", dest="can_notify",
                         default=False, help="sends an email on command completion")

    optparser.add_option("--tags", dest="tags",
                         help="comma-separated list of tags to be associated with the query ( e.g., tag1 tag1,tag2 )")

    optparser.add_option("--name", dest="name",
                         help="Assign a name to this command")

    optparser.add_option("--print-logs", action="store_true", dest="print_logs",
                         default=False, help="Fetch logs and print them to stderr.")

    @classmethod
    def parse(cls, args):
        """
        Parse command line arguments to construct a dictionary of command
        parameters that can be used to create a command

        Args:
            `args`: sequence of arguments

        Returns:
            Dictionary that can be used in create method

        Raises:
            ParseError: when the arguments are not correct
        """

        try:
            (options, args) = cls.optparser.parse_args(args)
            if options.mode not in ["1", "2"]:
                raise ParseError("mode must be either '1' or '2'",
                                 cls.optparser.format_help())

            if (options.dbtap_id is None) or (options.db_table is None):
                raise ParseError("dbtap_id and db_table are required",
                                 cls.optparser.format_help())

            if options.mode is "1":
                if options.hive_table is None:
                    raise ParseError("hive_table is required for mode 1",
                                     cls.optparser.format_help())
            elif options.export_dir is None:    # mode 2
                raise ParseError("export_dir is required for mode 2",
                                 cls.optparser.format_help())

            if options.db_update_mode is not None:
                if options.db_update_mode not in ["allowinsert", "updateonly"]:
                    raise ParseError("db_update_mode should either be left blank for append "
                                     "mode or be 'updateonly' or 'allowinsert'",
                                     cls.optparser.format_help())
                if options.db_update_mode is "updateonly":
                    if options.db_update_keys is None:
                        raise ParseError("db_update_keys is required when db_update_mode "
                                         "is 'updateonly'",
                                         cls.optparser.format_help())
                elif options.db_update_keys is not None:
                    raise ParseError("db_update_keys is used only when db_update_mode "
                                     "is 'updateonly'",
                                     cls.optparser.format_help())

        except OptionParsingError as e:
            raise ParseError(e.msg, cls.optparser.format_help())
        except OptionParsingExit as e:
            return None

        v = vars(options)
        v["command_type"] = "DbExportCommand"
        return v


class DbexportCommand(DbExportCommand):
    pass


class DbImportCommand(Command):
    usage = "dbimportcmd <submit|run> [options]"

    optparser = GentleOptionParser(usage=usage)
    optparser.add_option("-m", "--mode", dest="mode",
                         help="Can be 1 for Hive export or 2 for HDFS/S3 export")
    optparser.add_option("--hive_table", dest="hive_table",
                         help="Mode 1: Name of the Hive Table from which data will be exported")
    optparser.add_option("--dbtap_id", dest="dbtap_id",
                         help="Modes 1 and 2: DbTap Id of the target database in Qubole")
    optparser.add_option("--db_table", dest="db_table",
                         help="Modes 1 and 2: Table to export to in the target database")
    optparser.add_option("--where_clause", dest="db_where",
                         help="Mode 1: where clause to be applied to the table before extracting rows to be imported")
    optparser.add_option("--parallelism", dest="db_parallelism",
                         help="Mode 1 and 2: Number of parallel threads to use for extracting data")

    optparser.add_option("--extract_query", dest="db_extract_query",
                         help="Modes 2: SQL query to be applied at the source database for extracting data. "
                              "$CONDITIONS must be part of the where clause")
    optparser.add_option("--boundary_query", dest="db_boundary_query",
                         help="Mode 2: query to be used get range of rowids to be extracted")
    optparser.add_option("--split_column", dest="db_split_column",
                         help="column used as rowid to split data into range")

    optparser.add_option("--notify", action="store_true", dest="can_notify",
                         default=False, help="sends an email on command completion")

    optparser.add_option("--tags", dest="tags",
                         help="comma-separated list of tags to be associated with the query ( e.g., tag1 tag1,tag2 )")

    optparser.add_option("--name", dest="name",
                         help="Assign a name to this command")

    optparser.add_option("--print-logs", action="store_true", dest="print_logs",
                         default=False, help="Fetch logs and print them to stderr.")

    @classmethod
    def parse(cls, args):
        """
        Parse command line arguments to construct a dictionary of command
        parameters that can be used to create a command

        Args:
            `args`: sequence of arguments

        Returns:
            Dictionary that can be used in create method

        Raises:
            ParseError: when the arguments are not correct
        """

        try:
            (options, args) = cls.optparser.parse_args(args)
            if options.mode not in ["1", "2"]:
                raise ParseError("mode must be either '1' or '2'",
                                 cls.optparser.format_help())

            if (options.dbtap_id is None) or (options.db_table is None):
                raise ParseError("dbtap_id and db_table are required",
                                 cls.optparser.format_help())

            # TODO: Semantic checks for parameters in mode 1 and 2

        except OptionParsingError as e:
            raise ParseError(e.msg, cls.optparser.format_help())
        except OptionParsingExit as e:
            return None

        v = vars(options)
        v["command_type"] = "DbImportCommand"
        return v


class CompositeCommand(Command):
    @classmethod
    def compose(cls, sub_commands, macros=None, cluster_label=None, notify=False, name=None, tags=None):
        """
        Args:
            `sub_commands`: list of sub-command dicts

        Returns:
            Dictionary that can be used in create method

        Example Usage:
            cmd1 = HiveCommand.parse(['--query', "show tables"])
            cmd2 = PigCommand.parse(['--script_location', "s3://paid-qubole/PigAPIDemo/scripts/script1-hadoop-s3-small.pig"])
            composite = CompositeCommand.compose([cmd1, cmd2])
            cmd = CompositeCommand.run(**composite)
        """
        if macros is not None:
            macros = json.loads(macros)
        return {
                "sub_commands": sub_commands,
                "command_type": "CompositeCommand",
                "macros": macros,
                "label": cluster_label,
                "tags": tags,
                "can_notify": notify,
                "name": name
               }


class DbTapQueryCommand(Command):
    usage = "dbtapquerycmd <submit|run> [options]"

    optparser = GentleOptionParser(usage=usage)
    optparser.add_option("--db_tap_id", dest="db_tap_id",
                         help="dbTap Id of the target database in Qubole")
    optparser.add_option("-q", "--query", dest="query", help="query string")
    optparser.add_option("--notify", action="store_true", dest="can_notify",
                         default=False, help="sends an email on command completion")
    optparser.add_option("--macros", dest="macros",
                         help="expressions to expand macros used in query")

    optparser.add_option("--tags", dest="tags",
                         help="comma-separated list of tags to be associated with the query ( e.g., tag1 tag1,tag2 )")
    optparser.add_option("--name", dest="name",
                         help="Assign a name to this command")

    optparser.add_option("--print-logs", action="store_true", dest="print_logs",
                         default=False, help="Fetch logs and print them to stderr.")

    @classmethod
    def parse(cls, args):
        """
        Parse command line arguments to construct a dictionary of command
        parameters that can be used to create a command

        Args:
            `args`: sequence of arguments

        Returns:
            Dictionary that can be used in create method

        Raises:
            ParseError: when the arguments are not correct
        """

        try:
            (options, args) = cls.optparser.parse_args(args)
            if (options.db_tap_id is None):
                raise ParseError("db_tap_id is required",
                                 cls.optparser.format_help())
            if (options.query is None):
                raise ParseError("query is required",
                                 cls.optparser.format_help())

        except OptionParsingError as e:
            raise ParseError(e.msg, cls.optparser.format_help())
        except OptionParsingExit as e:
            return None

        if options.macros is not None:
            options.macros = json.loads(options.macros)
        v = vars(options)
        v["command_type"] = "DbTapQueryCommand"
        return v

def _read_iteratively(key_instance, fp, delim):
    key_instance.open_read()
    while True:
        try:
            # Default buffer size is 8192 bytes
            data = next(key_instance)
            if sys.version_info < (3, 0, 0):
                fp.write(str(data).replace(chr(1), delim))
            else:
                import io
                if isinstance(fp, io.TextIOBase):
                    fp.buffer.write(data.decode('utf-8').replace(chr(1), delim).encode('utf8'))
                elif isinstance(fp, io.BufferedIOBase) or isinstance(fp, io.RawIOBase):
                    fp.write(data.decode('utf8').replace(chr(1), delim).encode('utf8'))
                else:
                    # Can this happen? Don't know what's the right thing to do in this case.
                    pass
        except StopIteration:
            # Stream closes itself when the exception is raised
            return


def _download_to_local(boto_conn, s3_path, fp, num_result_dir, delim=None):
    '''
    Downloads the contents of all objects in s3_path into fp

    Args:
        `boto_conn`: S3 connection object

        `s3_path`: S3 path to be downloaded

        `fp`: The file object where data is to be downloaded
    '''
    #Progress bar to display download progress
    def _callback(downloaded, total):
        '''
        Call function for upload.

        `downloaded`: File size already downloaded (int)

        `total`: Total file size to be downloaded (int)
        '''
        if (total is 0) or (downloaded == total):
            return
        progress = downloaded*100/total
        sys.stderr.write('\r[{0}] {1}%'.format('#'*progress, progress))
        sys.stderr.flush()

    def _is_complete_data_available(bucket_paths, num_result_dir):
        if num_result_dir == -1:
            return True
        unique_paths = set()
        files = {}
        for one_path in bucket_paths:
            name = one_path.name.replace(key_prefix, "", 1)
            if name.startswith('_tmp.'):
                continue
            path = name.split("/")
            dir = path[0].replace("_$folder$", "", 1)
            unique_paths.add(dir)
            if len(path) > 1:
                file = int(path[1])
                if dir not in files:
                    files[dir] = []
                files[dir].append(file)
        if len(unique_paths) < num_result_dir:
            return False
        for k in files:
            v = files.get(k)
            if len(v) > 0 and max(v) + 1 > len(v):
                return False
        return True

    m = _URI_RE.match(s3_path)
    bucket_name = m.group(1)
    bucket = boto_conn.get_bucket(bucket_name)
    retries = 6
    if s3_path.endswith('/') is False:
        #It is a file
        key_name = m.group(2)
        key_instance = bucket.get_key(key_name)
        while key_instance is None and retries > 0:
            retries = retries - 1
            log.info("Results file is not available on s3. Retry: " + str(6-retries))
            time.sleep(10)
            key_instance = bucket.get_key(key_name)
        if key_instance is None:
            raise Exception("Results file not available on s3 yet. This can be because of s3 eventual consistency issues.")
        log.info("Downloading file from %s" % s3_path)
        if delim is None:
            key_instance.get_contents_to_file(fp)  # cb=_callback
        else:
            # Get contents as string. Replace parameters and write to file.
            _read_iteratively(key_instance, fp, delim=delim)

    else:
        #It is a folder
        key_prefix = m.group(2)
        bucket_paths = bucket.list(key_prefix)
        complete_data_available = _is_complete_data_available(bucket_paths, num_result_dir)
        while complete_data_available is False and retries > 0:
            retries = retries - 1
            log.info("Results dir is not available on s3. Retry: " + str(6-retries))
            time.sleep(10)
            complete_data_available = _is_complete_data_available(bucket_paths, num_result_dir)
        if complete_data_available is False:
            raise Exception("Results file not available on s3 yet. This can be because of s3 eventual consistency issues.")

        for one_path in bucket_paths:
            name = one_path.name

            # Eliminate _tmp_ files which ends with $folder$
            if name.endswith('$folder$'):
                continue

            log.info("Downloading file from %s" % name)
            if delim is None:
                one_path.get_contents_to_file(fp)  # cb=_callback
            else:
                _read_iteratively(one_path, fp, delim=delim)
